# Agilent_VNA_E5071C driver, P. Macha, modified by M. Weides July 2013, J. Braumueller 2015
# Adapted to Keysight VNA by A. Schneider and L. Gurenhaupt 2016
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
from time import sleep
import numpy

class Keysight_VNA_E5071C(Instrument):
    '''
    This is the python driver for the Anritsu MS4642A Vector Network Analyzer

    Usage:
    Initialise with
    <name> = instruments.create('<name>', address='<GPIB address>', reset=<bool>)
    
    '''

    def __init__(self, name, address, channel_index = 1):
        '''
        Initializes 

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
        '''
        
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address)
        if visa.qkit_visa_version > 1:
            # we have to define the read termination chars manually for the newer version
            idn = self._visainstrument.query('*IDN?')
            self._visainstrument.read_termination = idn[len(idn.strip()):]
            
        self._zerospan = False
        self._freqpoints = 0
        self._ci = channel_index
        self._pi = 2 # port_index, similar to self._ci
        self._active_trace = 0
        self._start = 0
        self._stop = 0
        self._nop = 0

        # Implement parameters
        self.add_parameter('nop', type=int,
            flags=Instrument.FLAG_GETSET,
            minval=2, maxval=20001,
            tags=['sweep'])
            
        self.add_parameter('bandwidth',     type=float, minval=0,   maxval=1e9, units='Hz')
        self.add_parameter('averages',      type=int,   minval=1,   maxval=1024)
        self.add_parameter('Average',       type=bool)
        self.add_parameter('centerfreq',    type=float, minval=0, maxval=20e9,  units='Hz')
        self.add_parameter('cwfreq',        type=float, minval=0, maxval=20e9,  units='Hz')
        self.add_parameter('startfreq',     type=float, minval=0, maxval=20e9,  units='Hz')
        self.add_parameter('stopfreq',      type=float, minval=0, maxval=20e9,  units='Hz')
        self.add_parameter('span',          type=float, minval=0, maxval=20e9,  units='Hz')
        self.add_parameter('power',         type=float, minval=-85, maxval=10,  units='dBm')
        self.add_parameter('startpower',    type=float, minval=-85, maxval=10,  units='dBm')
        self.add_parameter('stoppower',     type=float, minval=-85, maxval=10,  units='dBm')
        self.add_parameter('cw',            type=bool)
        self.add_parameter('zerospan',      type=bool)
        self.add_parameter('channel_index', type=int)
        self.add_parameter('sweeptime',     type=float, minval=0, maxval=1e3,   units='s',flags=Instrument.FLAG_GET)
        self.add_parameter('sweeptime_averages', type=float,minval=0, maxval=1e3,units='s',flags=Instrument.FLAG_GET)
        self.add_parameter('edel',          type=float,minval=-10, maxval=10,units='s',channels=(1, self._pi), channel_prefix = 'port%d_') # the channel option for qtlab's Instument class is used here to easily address the two VNA ports
        self.add_parameter('edel_status',   type=bool) # legacy name for parameter. This corresponds to the VNA's port extension values.
        self.add_parameter('sweep_mode',    type=str)  #JDB This parameter switches on/off hold. The hold function below does the same job, this is just for code compatibility to the agilent and anritsu drivers.
        self.add_parameter('sweep_type',    type=str)
        self.add_parameter('active_trace',  type=int)
        
        #Triggering Stuff
        self.add_parameter('trigger_source', type=str)
        
        # Implement functions
        self.add_function('get_freqpoints')
        self.add_function('get_tracedata')
        self.add_function('init')
        self.add_function('avg_clear')
        self.add_function('avg_status')
        self.add_function('def_trig')
        self.add_function('get_hold')
        self.add_function('hold')
        self.add_function('get_sweeptime')
        self.add_function('get_sweeptime_averages')
        self.add_function('pre_measurement')
        self.add_function('start_measurement')
        self.add_function('ready')
        self.add_function('post_measurement')

        self.get_all()
    
    def get_all(self):
        self.get_nop()
        self.get_power()
        self.get_centerfreq()
        self.get_startfreq()
        self.get_stopfreq()
        self.get_span()
        self.get_bandwidth()
        self.get_trigger_source()
        self.get_Average()
        self.get_averages()
        self.get_freqpoints()
        self.get_channel_index()
        self.get_zerospan()
        self.get_sweeptime()
        self.get_sweeptime_averages()
        self.get_edel_status()
        self.get_cw()
        self.get_cwfreq()
        for port in range(self._pi):
            self.get('port%d_edel' % (port+1))
        
    ###
    #Communication with device
    ###
    
    
    def hold(self, status):
        self.write(":TRIG:SOUR INT")
        if status:
            self.write(':INIT%i:CONT OFF'%(self._ci))
        else:
            self.write(':INIT%i:CONT ON'%(self._ci))

    def get_hold(self):
        return self.ask(':INIT%i:CONT?'%(self._ci))
    
    def init(self):
        if self._zerospan:
          self.write('INIT1;*wai')
        else:
          if self.get_Average():
            for i in range(self.get_averages()):
              self.write('INIT1;*wai')
          else:
              self.write('INIT1;*wai')

    def def_trig(self):
        self.write(':TRIG:AVER ON')
        self.write(':TRIG:SOUR bus')
        
    def avg_clear(self):
        self.write(':SENS%i:AVER:CLE' %(self._ci))

    def avg_status(self):
        return 0 == (int(self.ask('STAT:OPER:COND?')) & (1<<4))
        
    def get_tracedata(self, format = 'AmpPha', single=False, averages=1.):
        '''
        Get the data of the current trace

        Input:
            format (string) : 'AmpPha': Amp in dB and Phase, 'RealImag',

        Output:
            'AmpPha':_ Amplitude and Phase
        '''
        
        if single==True:
            #print('single shot readout')
            self.write('TRIG:SOUR INT') #added MW July 2013. start single sweep.
            self.write('INIT%i:CONT ON'%(self._ci)) #added MW July 2013. start single sweep.
            self.hold(True)
            sleep(float(self.ask('SENS1:SWE:TIME?')))
        
        #sleep(0.1) # required to avoid timing issues    MW August 2013   ???
        
        self.write('FORM:DATA REAL')
        self.write('FORM:BORD SWAPPED')
        data = self.ask_for_values('CALC%i:SEL:DATA:SDAT?'%(self._ci), fmt = 3)
        data_size = numpy.size(data)
        datareal = numpy.array(data[0:data_size:2])
        dataimag = numpy.array(data[1:data_size:2])
        
        if format == 'RealImag':
          if self._zerospan:
            return numpy.mean(datareal), numpy.mean(dataimag)
          else:
            return datareal, dataimag
        elif format == 'AmpPha':
          if self._zerospan or self.get_cw(False):
            datacomplex = [numpy.mean(datareal + 1j*dataimag)]
            dataamp = numpy.abs(datacomplex)
            datapha = numpy.angle(datacomplex)
          else:
            dataamp = numpy.sqrt(datareal*datareal+dataimag*dataimag)
            datapha = numpy.arctan2(dataimag,datareal)
          return dataamp, datapha
        else:
          raise ValueError('get_tracedata(): Format must be AmpPha or RealImag')
      
    def get_freqpoints(self, query = False):
      if query:
           self.write("FORM:DATA REAL; FORM:BORD SWAPPED;")
           self._freqpoints = self.ask_for_values(':SENS%i:FREQ:DATA?'%(self._ci), format = visa.double)
      elif self.get_cw():
           self._freqpoints = numpy.atleast_1d(self.get_cwfreq())
      else:
           self._freqpoints = numpy.linspace(self._start,self._stop,self._nop)
      return self._freqpoints

    ###
    # SET and GET functions
    ###
    
    def do_set_sweep_mode(self, mode):
        '''
        select the sweep mode from 'hold', 'cont', single'
        single means only one single trace, not all the averages even if averages
         larger than 1 and Average==True
        '''
        if mode == 'hold':
            self.write(':INIT%i:CONT OFF'%(self._ci))
        elif mode == 'cont':
            self.write(':INIT%i:CONT ON'%(self._ci))
        elif mode == 'single':
            self.write(':INIT%i:CONT ON'%(self._ci))
            self.write(':INIT%i:CONT OFF'%(self._ci))
        else:
            logging.warning('invalid mode')
            
    def do_get_sweep_mode(self):
        return int(self.ask(':INIT%i:CONT?'%(self._ci)))
    
    def do_set_nop(self, nop):
        '''
        Set Number of Points (nop) for sweep

        Input:
            nop (int) : Number of Points

        Output:
            None
        '''

        logging.debug(__name__ + ' : setting Number of Points to %s ' % (nop))
        if self._zerospan:
          print('in zerospan mode, nop is 1')
        else:
          cw = self.get_cw()
          if(nop == 1) and (not cw):
              logging.info(__name__ + 'nop == 1 is only supported in CW mode.')
              self.set_cw(True)
          else:
              self.write(':SENS%i:SWE:POIN %i' %(self._ci,nop))
          """
          Comment below is copy from Anritsu driver
          if(cw):
            self.write(':SENS%i:SWE:CW:POIN %i' %(self._ci,nop))
          else:
            self.write(':SENS%i:SWE:POIN %i' %(self._ci,nop))
          """
          self._nop = nop
          self.get_freqpoints() #Update List of frequency points
        
    def do_get_nop(self):
        '''
        Get Number of Points (nop) for sweep

        Input:
            None
        Output:
            nop (int)
        '''
        logging.debug(__name__ + ' : getting Number of Points')
        if self._zerospan or self.get_cw():
          self._nop = 1
        else:
          self._nop = int(self.ask(':SENS%i:SWE:POIN?' %(self._ci)))
        return self._nop
    
    def do_set_Average(self, status):
        '''
        Set status of Average

        Input:
            status (boolean)

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting Average to "%s"' % (status))
        if status:
            status = 'ON'
            self.write('SENS%i:AVER:STAT %s' % (self._ci,status))
        elif status == False:
            status = 'OFF'
            self.write('SENS%i:AVER:STAT %s' % (self._ci,status))
        else:
            raise ValueError('set_Average(): can only set True or False')
    def do_get_Average(self):
        '''
        Get status of Average

        Input:
            None

        Output:
            Status of Averaging (boolean)
        '''
        logging.debug(__name__ + ' : getting average status')
        return bool(int(self.ask('SENS%i:AVER:STAT?' %(self._ci))))
        
    def do_set_averages(self, av):
        '''
        Set number of averages

        Input:
            av (int) : Number of averages

        Output:
            None
        '''
        if self._zerospan == False:
            logging.debug(__name__ + ' : setting Number of averages to %i ' % (av))
            self.write('SENS%i:AVER:COUN %i' % (self._ci,av))
        else:
            self.write('SWE%i:POIN %.1f' % (self._ci,av))
            
    def do_get_averages(self):
        '''
        Get number of averages

        Input:
            None
        Output:
            number of averages
        '''
        logging.debug(__name__ + ' : getting Number of Averages')
        if self._zerospan:
          return int(self.ask('SWE%i:POIN?' % self._ci))
        else:
          return int(self.ask('SENS%i:AVER:COUN?' % self._ci))
          
    def do_set_power(self,pow):
        '''
        Set probe power

        Input:
            pow (float) : Power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting power to %s dBm' % pow)
        if self.get_cw():
            self.set_startpower(pow)
            self.set_stoppower(pow)
        else:
            self.write('SOUR%i:POW:PORT1:LEV:IMM:AMPL %.1f' % (self._ci,pow))
    def do_get_power(self):
        '''
        Get probe power

        Input:
            None

        Output:
            pow (float) : Power in dBm
        '''
        logging.debug(__name__ + ' : getting power')
        if self.get_cw():
            return self.get_startpower()
        else:
            return float(self.ask('SOUR%i:POW:PORT1:LEV:IMM:AMPL?' % (self._ci)))

    def do_set_startpower(self,pow):
        '''
        Set probe power

        Input:
            pow (float) : Power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting startpower to %s dBm' % pow)
        self.write('SOUR%i:POW:START %.1f' % (self._ci,pow))
    def do_get_startpower(self):
        '''
        Get probe power

        Input:
            None

        Output:
            pow (float) : Power in dBm
        '''
        logging.debug(__name__ + ' : getting startpower')
        return float(self.ask('SOUR%i:POW:START?' % (self._ci)))

    def do_set_stoppower(self,pow):
        '''
        Set probe power

        Input:
            pow (float) : Power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting stoppower to %s dBm' % pow)
        self.write('SOUR%i:POW:STOP %.1f' % (self._ci,pow))
    def do_get_stoppower(self):
        '''
        Get probe power

        Input:
            None

        Output:
            pow (float) : Power in dBm
        '''
        logging.debug(__name__ + ' : getting stoppower')
        return float(self.ask('SOUR%i:POW:STOP?' % (self._ci)))
        
    def do_set_centerfreq(self,cf):
        '''
        Set the center frequency

        Input:
            cf (float) :Center Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting center frequency to %s' % cf)
        if self.get_cw(False):
          self.set_cwfreq(cf)
        self.write('SENS%i:FREQ:CENT %f' % (self._ci,cf))
        self.get_startfreq()
        self.get_stopfreq()
        self.get_span()
    def do_get_centerfreq(self):
        '''
        Get the center frequency

        Input:
            None

        Output:
            cf (float) :Center Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting center frequency')
        return  float(self.ask('SENS%i:FREQ:CENT?'%(self._ci)))

    def do_set_cwfreq(self, cf):
        ''' set cw frequency '''
        self.write('SENS%i:FREQ:CW %f' % (self._ci,cf))
    
    def do_get_cwfreq(self):
        ''' get cw frequency '''
        return float(self.ask('SENS%i:FREQ:CW?'%(self._ci)))
        
    def do_set_span(self,span):
        '''
        Set Span

        Input:
            span (float) : Span in KHz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting span to %s Hz' % span)
        self.write('SENS%i:FREQ:SPAN %i' % (self._ci,span))
        self.get_startfreq()
        self.get_stopfreq()
        self.get_centerfreq()
        
    def do_get_span(self):
        '''
        Get Span
        
        Input:
            None

        Output:
            span (float) : Span in Hz
        '''
        #logging.debug(__name__ + ' : getting center frequency')
        span = self.ask('SENS%i:FREQ:SPAN?' % (self._ci) ) #float( self.ask('SENS1:FREQ:SPAN?'))
        return span

    def do_get_sweeptime_averages(self):###JB
        '''
        Get sweeptime
        
        Input:
            None

        Output:
            sweep time (float) times number of averages: sec
        '''
        return self.get_sweeptime() * (self.get_averages() if self.get_Average() else 1)
        
    def do_get_sweeptime(self):  #added MW July 2013
        '''
        Get sweeptime
        
        Input:
            None

        Output:
            sweep time (float) : sec
        '''
        logging.debug(__name__ + ' : getting sweep time')
        self._sweep=float(self.ask('SENS1:SWE:TIME?'))
        return self._sweep


    def do_set_edel(self, val,channel):  # MP 04/2017

        '''
        Set electrical delay

        '''
        logging.debug(__name__ + ' : setting port %s extension to %s sec' % (channel, val))
        self.write('SENS1:CORR:EXT:PORT%i:TIME %.12f' % (channel, val))
        
    
    def do_get_edel(self, channel):   # MP 04/2017

        '''
        Get electrical delay

        '''
        logging.debug(__name__ + ' : getting port %s extension' % channel)
        self._edel = float(self.ask('SENS1:CORR:EXT:PORT%i:TIME?'% channel))
        return  self._edel
        
    def do_set_edel_status(self, status):   # AS 04/2019

        '''
        Set electrical delay

        '''
        logging.debug(__name__ + ' : setting port extension status to %s' % (status))
        self.write('SENS:CORR:EXT:STAT %i' % (status))
        
    
    def do_get_edel_status(self):   # AS 04/2019

        '''
        Get electrical delay

        '''
        logging.debug(__name__ + ' :  port extension status')
        return  self.ask('SENS:CORR:EXT:STAT?').strip() == "1"
        
        
    def do_set_startfreq(self,val):
        '''
        Set Start frequency

        Input:
            span (float) : Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting start freq to %s Hz' % val)
        self.write('SENS%i:FREQ:STAR %f' % (self._ci,val))
        self._start = val
        self.get_centerfreq()
        self.get_stopfreq()
        self.get_span()
        
    def do_get_startfreq(self):
        '''
        Get Start frequency
        
        Input:
            None

        Output:
            span (float) : Start Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting start frequency')
        self._start = float(self.ask('SENS%i:FREQ:STAR?' % (self._ci)))
        return  self._start

    def do_set_stopfreq(self,val):
        '''
        Set STop frequency

        Input:
            val (float) : Stop Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting stop freq to %s Hz' % val)
        self.write('SENS%i:FREQ:STOP %f' % (self._ci,val))
        self._stop = val
        self.get_startfreq()
        self.get_centerfreq()
        self.get_span()
    def do_get_stopfreq(self):
        '''
        Get Stop frequency
        
        Input:
            None

        Output:
            val (float) : Start Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting stop frequency')
        self._stop = float(self.ask('SENS%i:FREQ:STOP?' %(self._ci) ))
        return  self._stop

    def do_set_bandwidth(self,band):
        '''
        Set Bandwidth

        Input:
            band (float) : Bandwidth in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting bandwidth to %s Hz' % (band))
        self.write('SENS%i:BWID:RES %i' % (self._ci,band))
    def do_get_bandwidth(self):
        '''
        Get Bandwidth

        Input:
            None

        Output:
            band (float) : Bandwidth in Hz
        '''
        logging.debug(__name__ + ' : getting bandwidth')
        # getting value from instrument
        return  float(self.ask('SENS%i:BWID:RES?'%self._ci))
    
    def do_set_cw(self, val):
        '''
        Set instrument to CW (single frequency) mode and back
        '''
        if val:
            self.set_cwfreq(self.get_centerfreq())
            self.set_nop(2)
            self.set_startpower(self.get_power())
            self.set_stoppower(self.get_power())
            return self.write(':SENS%i:SWE:TYPE POW' %(self._ci))
        else:
            return self.write(':SENS%i:SWE:TYPE LIN' %(self._ci))
        
    def do_get_cw(self):
        '''
        retrieve CW mode status from device
        '''
        sweep_type = str(self.ask(':SENS%i:SWE:TYPE?'%(self._ci))).rstrip()
        if sweep_type == 'POW': ret = True
        else: ret = False
        return ret

    def do_set_zerospan(self,val):
        '''
        Zerospan is a virtual "zerospan" mode. In Zerospan physical span is set to
        the minimal possible value (2Hz) and "averages" number of points is set.

        Input:
            val (bool) : True or False

        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting status to "%s"' % status)
        if val not in [True, False]:
            raise ValueError('set_zerospan(): can only set True or False')
        if val:
          self._oldnop = self.get_nop()
          self._oldspan = self.get_span()
          if self.get_span() > 0.002:
            Warning('Setting ZVL span to 2Hz for zerospan mode')
            self.set_span(0.002)
            
        av = self.get_averages()
        self._zerospan = val
        if val:
            self.set_Average(False)
            self.set_averages(av)
            if av<2:
              av = 2
        else:
          self.set_Average(True)
          self.set_span(self._oldspan)
          self.set_nop(self._oldnop)
          self.get_averages()
        self.get_nop()
        
    def do_get_zerospan(self):
        '''
        Check weather the virtual zerospan mode is turned on

        Input:
            None

        Output:
            val (bool) : True or False
        '''
        return self._zerospan


    def do_set_trigger_source(self,source):
        '''
        Set Trigger Mode

        Input:
            source (string) : INTernal | MANual | EXTernal | REMote

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting trigger source to "%s"' % source)
        if source.upper() in ['INT', 'MAN', 'EXT', 'BUS']:
            self.write('TRIG:SEQ:SOUR %s' % source.upper())
        else:
            raise ValueError('set_trigger_source(): must be INTernal | MANual | EXTernal | REMote')

    def do_get_trigger_source(self):
        '''
        Get Trigger Mode

        Input:
            None

        Output:
            source (string) : INTernal | MANual | EXTernal | BUS
        '''
        logging.debug(__name__ + ' : getting trigger source')
        return str(self.ask('TRIG:SEQ:SOUR?')).rstrip()
        

    def do_set_channel_index(self,val):
        '''
        Set the index of the channel to address.

        Input:
            val (int) : 1 .. number of active channels (max 16)

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting channel index to "%i"' % int)
        nop = self._visainstrument.read('DISP:COUN?')
        if val < nop:
            self._ci = val
        else:
            raise ValueError('set_channel_index(): index must be < nop channels')
    def do_get_channel_index(self):
        '''
        Get active channel

        Input:
            None

        Output:
            channel_index (int) : 1-16
        '''
        logging.debug(__name__ + ' : getting channel index')
        return self._ci
    
    def do_set_active_trace(self, trace):
        """
        Sets the active trace, which can then be readout
        :param trace: Number of the active trace
        :return: None
        """
        # TODO: catch error
        self._active_trace = trace
        self.write('CALC{}:PAR{}:SEL'.format(self._ci, trace))

    def do_get_active_trace(self):
        """
        :return: the active trace on the VNA
        """
        # TODO: ask device
        return self._active_trace

    def do_get_sweep_type(self):
        '''
        Get the Sweep Type

        Input:
            None

        Output:
            Sweep Type (string). One of
            LIN:    Frequency-based linear sweep
            LOG:    Frequency-based logarithmic sweep
            SEGM:   Segment-based sweep with frequency-based segments
            POW:    Power-based sweep with either Power-based sweep with a 
                CW frequency, or Power-based sweep with swept-frequency
        '''
        logging.debug(__name__ + ' : getting sweep type')
        
        return str(self.ask('SENS%i:SWE:TYPE?' %(self._ci))).rstrip()
    
    def do_set_sweep_type(self,swtype):
        '''
        Set the Sweep Type
        Input:
            swtype (string):    One of
                LIN:    Frequency-based linear sweep
                LOG:    Frequency-based logarithmic sweep
                SEGM:   Segment-based sweep with frequency-based segments
                POW:    Power-based sweep with either Power-based sweep with a 
                    CW frequency, or Power-based sweep with swept-frequency

        Output:
            None            
        '''
        if swtype in ('LIN','LOG','SEGM','POW'):
            
            logging.debug(__name__ + ' : Setting sweep type to %s'%(swtype))
            return self.write('SENS%i:SWE:TYPE %s' %(self._ci,swtype))
            
        else:
            logging.debug(__name__ + ' : Illegal argument %s'%(swtype))
    
    def write(self,msg):
        return self._visainstrument.write(msg)
    
    if qkit.visa.qkit_visa_version == 1:
        def ask(self, msg):
            return self._visainstrument.ask(msg)
    
        def ask_for_values(self, msg, **kwargs):
            return self._visainstrument.ask_for_values(kwargs)
    else:
        def ask(self, msg):
            return self._visainstrument.query(msg)
    
        def ask_for_values(self, msg, format=None, fmt=None):
            dtype = format if format is not None else fmt if fmt is not None else qkit.visa.single
            dtype = qkit.visa.dtypes[dtype]
            return self._visainstrument.query_binary_values(msg,datatype=dtype,container=numpy.array)
     
    def pre_measurement(self):
        '''
        Set everything needed for the measurement
        '''
        self.write(":TRIG:SOUR BUS")#Only wait for software triggers
        self.write(":TRIG:AVER ON")# Tell the instrument to do the specific number of averages on every trigger.
        
        
    def post_measurement(self):
        '''
        After a measurement, the VNA is in hold mode, and it can be difficult to start a measurement again from front panel.
        This function brings the VNA back to normal measuring operation.
        '''
        self.write(":TRIG:SOUR INT") #Only wait for software triggers
        self.write(":TRIG:AVER OFF")# Tell the instrument to do the specific number of averages on every trigger.
        self.hold(False)
        
      
    def start_measurement(self):
        '''
        This function is called at the beginning of each single measurement in the spectroscopy script.
        Here, it resets the averaging
        '''
        self.avg_clear()
        self.write('*TRG') #go

    
    def ready(self):
        '''
        This is a proxy function, returning True when the VNA has finished the required number of averages.
        '''
        return (int(self.ask(':STAT:OPER:COND?')) & 32)==32



# MG@KIT 04/2022
"""
qkit instrument driver for qdevil QDAC-II.
 +- 24 channel low noise voltage source.
"""

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import logging

import qkit
from qkit.core.instrument_base import Instrument
from qkit import visa


class QDAC_II(Instrument):
    """
    Instrument class for QDevil QDAC-II.


    """

    def __init__(self, name, address, port=5025):
        """
        Initialize qkit paramters and connect to the device.
        """

        logging.info(__name__ + " : Initializing Instrument")
        Instrument.__init__(self, name, tags=["physical"])

        self._address = address
        self._port = port
        try:
            self._visainstrument = visa.instrument(
                "TCPIP::{:s}::{:i}::SOCKET".format(self._address, self._port)
            )
        except Exception as detail:
            logging.error(
                "Connection to QDvil Instrument not possible. Original error: {:s}".format(
                    detail
                )
            )
        self._visainstrument.read_termination = "\n"
        self._visainstrument.write_termination = "\n"

        # initial variables
        self._sweep_mode = 1  # IV-mode
        self._sweep_channels = (1,)
        self._measurement_modes = {0: "2-wire", 1: "4-wire"}
        self._IV_modes = {0: "curr", 1: "volt"}
        self._IV_units = {0: "A", 1: "V"}

    def write(self, cmd):
        """
        Sends a visa command <cmd>, waits until "operation complete" and raises eventual errors of the device.

        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.

        Returns
        -------
        None
        """
        self._visainstrument.write(cmd)
        while not bool(int(self._visainstrument.query("*OPC?"))):
            time.sleep(1e-6)
        self._raise_error()
        return

    def query(self, cmd):
        """
        Sends a visa command <cmd>, waits until "operation complete", raises eventual errors of the Device and returns the read answer <ans>.

        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.

        Returns
        -------
        answer: str
            Answer that is returned at query after the sent <cmd>.
        """
        if "?" in cmd:
            ans = self._visainstrument.query(cmd).rstrip()
        else:
            ans = self._visainstrument.query("{:s}?".format(cmd)).rstrip()
        while not bool(int(self._visainstrument.query("*opc?"))):
            time.sleep(1e-6)
        self._raise_error()
        return ans

    def get_OPC(self):
        """
        Waits until the condition of standard event register entry "Operation complete" is fulfilled.

        Parameters
        ----------
                None

        Returns
        -------
                None
        """
        try:
            logging.debug(
                '''{!s}: Get ccr event "Operation complete"'''.format(__name__)
            )
            return bool(int(self.query("*OPC")))
        except Exception as e:
            logging.error(
                '{!s}: Cannot get ccr condition "Operation complete"'.format(__name__)
            )
            raise type(e)(
                '{!s}: Cannot get ccr condition "Operation complete"\n{!s}'.format(
                    __name__, e
                )
            )

    def _wait_for_OPC(self):
        """
        Waits until the condition of standard event register entry "Operation complete" is fulfilled.

        Parameters
        ----------
                None

        Returns
        -------
                None
        """
        while not (self.get_OPC()):
            time.sleep(1e-3)
        return

    def clear_error(self):
        """
        Clears error of instrument.

        Parameters
        ----------
                None

        Returns
        -------
                None
        """
        try:
            logging.debug("{!s}: Clears error of instrument".format(__name__))
            self.write("*CLS")
        except Exception as e:
            logging.error("{!s}: Cannot clears error of instrument".format(__name__))
            raise type(e)(
                "{!s}: Cannot clears error of instrument\n{!s}".format(__name__, e)
            )
        return

    def reset(self):
        """
        Resets the instrument or a single channel to default conditions.

        Parameters
        ----------
        None

        Returns
        -------
                None
        """
        try:
            logging.debug("{!s}: Reset instrument to factory settings".format(__name__))
            self.write("*RST")
        except Exception as e:
            logging.error(
                "{!s}: Cannot reset instrument to factory settings".format(__name__)
            )
            raise type(e)(
                "{!s}: Cannot reset instrument to factory settings\n{!s}".format(
                    __name__, e
                )
            )
        return

    def get_error(self):
        """
        Gets errors of instrument and removes them from the error queue.

        Parameters
        ----------
                None

        Returns
        -------
        err: array of str
            Entries from the error queue
        """
        try:
            logging.debug("{!s}: Get errors of instrument".format(__name__))
            err = [self._visainstrument.query("syst:err?").rstrip().split(",", 1)]
            while err[-1] != ["0", '"No error"']:
                err.append(
                    self._visainstrument.query("syst:err?").rstrip().split(",", 1)
                )
            if len(err) > 1:
                err = err[:-1]
            err = [[int(e[0]), str(e[1][1:-1])] for e in err]
            return err
        except Exception as e:
            logging.error("{!s}: Cannot get errors of instrument".format(__name__))
            raise type(e)(
                "{!s}: Cannot get errors of instrument\n{!s}".format(__name__, e)
            )

    def _raise_error(self):
        """
        Gets errors of instrument and as the case may be raises a python error.

        Parameters
        ----------
                None

        Returns
        -------
                None
        """
        errors = self.get_error()
        if len(errors) == 1 and errors[0][0] is 0:  # no error
            return
        else:
            msg = __name__ + " raises the following errors:"
            for err in errors:
                msg += "\n{:s}: ({:s})".format(err[0], err[1])
            raise ValueError(msg)

    def clear_error(self):
        """
        Clears error of instrument.

        Parameters
        ----------
                None

        Returns
        -------
                None
        """
        try:
            logging.debug("{!s}: Clears error of instrument".format(__name__))
            self._write("*CLS")
        except Exception as e:
            logging.error("{!s}: Cannot clears error of instrument".format(__name__))
            raise type(e)(
                "{!s}: Cannot clears error of instrument\n{!s}".format(__name__, e)
            )
        return

    def close_connection(self):
        """
        Closes the VISA-instrument to disconnect the instrument.

        Parameters
        ----------
                None

        Returns
        -------
                None
        """
        try:
            logging.info(
                __name__
                + " : Closing connection to QDAC-II server {:s}".format(self._address)
            )
            self._visainstrument.close()
        except Exception as e:
            logging.error("{!s}: Cannot close VISA-instrument".format(__name__))
            raise type(e)(
                "{!s}: Cannot close VISA-instrument\n{!s}".format(__name__, e)
            )
        return

    def set_DC_voltage(channel, value, slew=2e7, **par):
        """
        Setter function for voltage in a channel.

        Parameter
        ---------
        channel: int or list of int.
            Channel(s) for which the voltage should be set.
        value: float.
            Voltage value in volts (V) that should be set in each channel.
        slew: float, default 2e7.
            Maximum voltage change rate in volts per second (V/s). Minimal value is 20
            and maximum is 2e7 (default).

        **par: dict type. optional.
            Including volt_range and lowpass_filter.
            volt_range: {None, 'LOW','HIGH'}
                Voltage range 'LOW' is up to 2 V and 'HIGH' up to 10 V. If not specified it
                will be calculated from the <value> parameter.
            lowpass_filter: {'DC', 'MED', 'HIGH'}
                'DC' filters out everything below ~10 Hz, 'MED' at 10 kHz and 'HIGH' at 100kHz.
                Default is 'DC'.

        Example
        -------
        q = QDAC_II("QDevil QDAC-II", 10.22.197.176)
        q.set_DC_voltage([1,2,6],0.2)
        """

        channel_str = ""
        for ch in channel:
            channel_str = channel_str + str(int(ch)) + ","

        try:
            logging.debug(
                "{!s}: Set voltage of channel(s) {:s} to {:d} with slew {:d}".format(
                    __name__, channel_str, value, slew
                )
            )
            self.write("sour:volt {:d}, (@{:s}) ".format(value, channel_str))
            self._set_DC_voltage_par(channel, value, slew, *par)
        except Exception as e:
            logging.error(
                "{!s}: Cannot set voltage of channel(s) {!s}".format(
                    __name__, channel_str
                )
            )
            raise type(e)(
                "{!s}: Cannot set voltage of channel(s) {!s}\n{!s}".format(
                    __name__, channel_str, e
                )
            )
        return

    def _set_DC_voltage_par(channel, value, slew=2e7, **par):
        """
        Private function for setting additional parameters for the DC voltage.

        Parameters
        ----------
        channel: int or list of int.
            Channel(s) for which the voltage should be set.
        value: float
            Voltage value you want to set. Used for computing voltage range, if no range
            is specified in par.
        slew: float, default 2e7.
            Maximum voltage change rate in volts per second (V/s). Minimal value is 20
            and maximum is 2e7 (default).

        **par: dict type. optional.
            Including volt_range and lowpass_filter.
            volt_range: {None, 'LOW','HIGH'}
                Voltage range 'LOW' is up to 2 V and 'HIGH' up to 10 V. If not specified it
                will be calculated from the <value> parameter.
            lowpass_filter: {'DC', 'MED', 'HIGH'}
                'DC' filters out everything below ~10 Hz, 'MED' at 10 kHz and 'HIGH' at 100kHz.
                Default is 'DC'.

        Returns
        -------
        None
        """

        channel_str = ""
        for ch in channel:
            channel_str = channel_str + str(int(ch)) + ","
        try:
            self.write("sour:volt:slew {:d}, (@{:s})".format(slew, channel_str))

            if volt_range not in par:
                if value < 2:
                    volt_range = "LOW"
                else:
                    volt_range = "HIGH"
            self.write("sour:volt:range {:s}, (@{:s})".format(volt_range, channel_str))
            if lowpass_filter in par:
                self.write(
                    "sour:volt:filter {:s}, (@{:s})".format(lowpass_filter, channel_str)
                )
            else:
                self.write("sour:volt:filter DC, (@{:s})".format(channel_str))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set voltage parameters of channel(s) {!s}".format(
                    __name__, channel_str
                )
            )
            raise type(e)(
                "{!s}: Cannot set voltage parameters of channel(s) {!s}\n{!s}".format(
                    __name__, channel_str, e
                )
            )
        return

    def get_current(channel):
        """
        Reads current from the device for given channels and returns result as a dict.

        Parameters
        ----------
        channel: int of list of int.
            Channel(s) for which the current should be read.

        Returns
        -------
        current: dict of floats.
            A dict containing the current in ampere (A) for each measured channel.

        Example
        -------
        q = QDAC_II("QDevil QDAC-II", 10.22.197.176)
        q.get_current([1,2,6])
        """
        self.query()  # TODO

    def take_IV(self, sweep):
        """
        Takes IV curve with sweep parameters <sweep> in set sweep mode.

        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using
            qkit.measure.transport.transport.sweep class)

        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        self.set_sweep_parameters(sweep=sweep)
        return self.get_tracedata()

    def set_sweep_paramters(self, sweep, dwell=1e-6, **par):
        """
        Sets sweep parameters <sweep> and prepares instrument for the set sweep mode.

        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using
            qkit.measure.transport.transport.sweep class)

        dwell: float.
            Duration of each level of the sweep in s. Must be beween 1e-6 (one sample at each level)
            and 3600s (10h). Default is 1e-6.
        **par: dict type. optional
            Dict of parameters that will be passed to set_DC_value. Containing settings for the
            voltage generator. Including lowpass_filter and volt_range.

        Returns
        -------
                None
        """

        if self._sweep_mode == 2:
            channel_bias, channel_sense = self._sweep_channels
        else:
            logging.error(
                "Mode {:d} not supported. This device only supports mode 2 (VI-mode).".format(
                    mode
                )
            )
            raise ValueError(
                "Mode {:d} not supported. This device only supports mode 2 (VI-mode).".format(
                    mode
                )
            )
        try:
            logging.debug(
                "{!s}: Set sweep parameters of channel {:d} and {:d} to {!s}".format(
                    __name__, channel_bias, channel_sense, sweep
                )
            )
            self._set_sweep_start(val=float(sweep[0]), channel=channel_bias)
            self._set_sweep_stop(val=float(sweep[1]), channel=channel_bias)
            self._set_sweep_step(val=np.abs(float(sweep[2])), channel=channel_bias)
            self._set_sweep_dwell(val=dwell, channel=channel_bias)
            # setup triggers
            self.set_trigger("int1", channel=channel_bias)
            # set int2 trigger to trigger on every step of the voltage sweep
            self.write("sour{:d}:dc:sst:tnum 2".format(channel_bias))
            self.set_sensor_trigger("int2", channel=channel_sense)

            self._set_DC_voltage_par(channel, value=sweep[1], **par)
            self._write("sour{:d}:mode swe".format(channel_bias))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set sweep parameters of channel {:d} and {:d} to {!s}".format(
                    __name__, channel_bias, channel_sense, sweep
                )
            )
            raise type(e)(
                "{!s}: Cannot set sweep parameters of channel {:d} and {:d} to {!s}\n{!s}".format(
                    __name__, channel_bias, channel_sense, sweep, e
                )
            )
        return

    def get_tracedata(self, sweep):
        """
        Starts bias sweep with already set parameters and gets trace data of bias <bias_values> and sense <sense_values> in the set sweep mode.

        Parameters
        ----------
                None

        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        if self._sweep_mode == 2:
            channel_bias, channel_sense = self._sweep_channels
        else:
            logging.error(
                "Mode {:d} not supported. This device only supports mode 2 (VI-mode).".format(
                    mode
                )
            )
            raise ValueError(
                "Mode {:d} not supported. This device only supports mode 2 (VI-mode).".format(
                    mode
                )
            )
        try:
            logging.debug(
                "{!s}: Take sweep data of bias channel {:d} and measurement channel{:d}".format(
                    __name__, channel_bias, channel_sense
                )
            )
            duration = self.query("sour{:d}:dc:swe:time?".format(channel_bias))
            self.write("sens{:d}:init:cont ON".format(channel_sense))
            self._wait_for_OPC()
            time.sleep(100e-6)
            self.write("tint1")
            time.sleep(duration)
            sense_values = np.fromstring(
                string=self.query("read? (@{:d})".format(channel_sense)),
                dtype=float,
                sep=",",
            )
            bias_values = get_sweep_bias_values(channel_bias)
            if len(bias_value) != len(sense_values):
                logging.error(
                    "Length of bias values ({:d}) does not match length of sensor values ({:d})".format(
                        len(bias_values), len(sense_values)
                    )
                )
                raise ValueError(
                    "Length of bias values ({:d}) does not match length of sensor values ({:d})".format(
                        len(bias_values), len(sense_values)
                    )
                )
            return bias_values, sense_values
        except Exception as e:
            logging.error(
                "{!s}: Cannot take sweep data of channel {!s} and {!s}".format(
                    __name__, channel_bias, channel_sense
                )
            )
            raise type(e)(
                "{!s}: Cannot take sweep data of channel {!s} and {!s}\n{!s}".format(
                    __name__, channel_bias, channel_sense, e
                )
            )

    def get_sweep_bias_values(self, channel):
        """
        Read out bias values of a sweep for a given channel.

        Parameters
        ----------
        channel: int
            Bias channel. Must be between 1 and 24.

        Returns
        -------
        values: list of float
            List of bias values
        """
        try:
            start = float(self.query("sour{:d}:swe:star?".format(channel)))
            stop = float(self.query("sour{:d}:swe:stop?".format(channel)))
            step = float(self.query("sour{:d}:swe:step?".foramt(channel)))
            return list(range(start, stop + step, step))
        except Exception as e:
            logging.error(
                "{:s}: Cannot get bias values of sweep of channels {:s}. Start, Stop or Stepsize not specified.".format(
                    __name__, channel
                )
            )
            raise type(e)(
                "{:s}: Cannot get bias values of sweep of channels {:s}. Start, Stop or Stepsize not specified.\n{:s}".format(
                    __name__, channel, e
                )
            )

    def set_trigger(self, trigger, channel=1):
        """
        Sets trigger of channel <channel> to <trigger>.

        Parameters
        ----------
        trigger: str
            Trigger to use. For example IMM, INT1 or EXT1.
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
                None
        """
        try:
            logging.debug(
                "{!s}: Set trigger of channel {:d} to {:s}".format(
                    __name__, channel, trigger
                )
            )
            self._write("sour{:d}:trig:sour {:s}".format(channel, trigger))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set trigger of channel {!s} to {!s}".format(
                    __name__, channel, trigger
                )
            )
            raise type(e)(
                "{!s}: Cannot set trigger of channel {!s} to {!s}\n{!s}".format(
                    __name__, channel, trigger, e
                )
            )
        return

    def get_trigger(self, channel=1):
        """
        Gets trigger of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        trigger: str
            Trigger of given channel. For example INT1.
        """
        try:
            logging.debug("{!s}: Get trigger of channel {:d}".format(__name__, channel))
            trigger = str(self.query("sour{:d}:trig:sourc?".format(channel)).lower())
            return trigger
        except Exception as e:
            logging.error(
                "{!s}: Cannot get trigger of channel {!s}".format(__name__, channel)
            )
            raise type(e)(
                "{!s}: Cannot get trigger of channel {!s}\n{!s}".format(
                    __name__, channel, e
                )
            )

    def set_sensor_trigger(self, trigger, channel=1):
        """
        Sets trigger for the current sensor of channel <channel> to <trigger>.

        Parameters
        ----------
        trigger: str
            Trigger to use. For example IMM, INT1 or EXT1.
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
                None
        """
        try:
            logging.debug(
                "{!s}: Set sensor trigger of channel {:d} to {:s}".format(
                    __name__, channel, trigger
                )
            )
            self._write("sens{:d}:trig:sour {:s}".format(channel, trigger))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set sensor trigger of channel {!s} to {!s}".format(
                    __name__, channel, trigger
                )
            )
            raise type(e)(
                "{!s}: Cannot set sensor trigger of channel {!s} to {!s}\n{!s}".format(
                    __name__, channel, trigger, e
                )
            )
        return

    def get_sensor_trigger(self, channel=1):
        """
        Gets trigger for the current sensor of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        trigger: str
            Trigger of given channel. For example INT1.
        """
        try:
            logging.debug(
                "{!s}: Get sensor trigger of channel {:d}".format(__name__, channel)
            )
            trigger = str(self.query("sens{:d}:trig:sourc?".format(channel)).lower())
            return trigger
        except Exception as e:
            logging.error(
                "{!s}: Cannot get sensor trigger of channel {!s}".format(
                    __name__, channel
                )
            )
            raise type(e)(
                "{!s}: Cannot get sensor trigger of channel {!s}\n{!s}".format(
                    __name__, channel, e
                )
            )

    def _set_sweep_start(self, val, channel):
        """
        Sets sweep start value of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Start value of sweep.
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
                None
        """
        try:
            logging.debug(
                "{!s}: Set sweep start of channel {:d} to {:f}".format(
                    __name__, channel, val
                )
            )
            self.write("sour{:d}:swe:star {:f}".format(channel, val))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set sweep start of channel {!s} to {!s}".format(
                    __name__, channel, val
                )
            )
            raise type(e)(
                "{!s}: Cannot set sweep start of channel {!s} to {!s}\n{!s}".format(
                    __name__, channel, val, e
                )
            )
        return

    def _get_sweep_start(self, channel=1):
        """
        Gets sweep start value of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        val: float
            Start value of sweep.
        """
        try:
            logging.debug(
                "{!s}: Get sweep start of channel {:d}".format(__name__, channel)
            )
            return float(self.query(":sour{:d}:swe:star".format(channel)))
        except Exception as e:
            logging.error(
                "{!s}: Cannot get sweep start of channel {!s}".format(__name__, channel)
            )
            raise type(e)(
                "{!s}: Cannot get sweep start of channel {!s}\n{!s}".format(
                    __name__, channel, e
                )
            )

    def set_sweep_dwell(self, val, channel):
        """
        Sets sweep dwell value of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Dwell value of sweep. The duration of each level in the sweep. Must be between
            1e-6 (1 sample) and 3600s (10h).
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
                None
        """
        try:
            logging.debug(
                "{!s}: Set sweep dwell of channel {:d} to {:f}".format(
                    __name__, channel, val
                )
            )
            self.write("sour{:d}:swe:dweli {:f}".format(channel, val))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set sweep dwell of channel {!s} to {!s}".format(
                    __name__, channel, val
                )
            )
            raise type(e)(
                "{!s}: Cannot set sweep dwell of channel {!s} to {!s}\n{!s}".format(
                    __name__, channel, val, e
                )
            )
        return

    def get_sweep_dwell(self, channel=1):
        """
        Gets sweep dwell value of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        val: float
            Dwell value of sweep. The duration of each level in the sweep in seconds.
        """
        try:
            logging.debug(
                "{!s}: Get sweep dwell of channel {:d}".format(__name__, channel)
            )
            return float(self.query(":sour{:d}:swe:dweli".format(channel)))
        except Exception as e:
            logging.error(
                "{!s}: Cannot get sweep dwell of channel {!s}".format(__name__, channel)
            )
            raise type(e)(
                "{!s}: Cannot get sweep dwell of channel {!s}\n{!s}".format(
                    __name__, channel, e
                )
            )

    def _set_sweep_stop(self, val, channel):
        """
        Sets sweep stop value of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Stop value of sweep.
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
                None
        """
        try:
            logging.debug(
                "{!s}: Set sweep stop of channel {:d} to {:f}".format(
                    __name__, channel, val
                )
            )
            self.write("sour{:d}:swe:stop {:f}".format(channel, val))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set sweep stop of channel {!s} to {!s}".format(
                    __name__, channel, val
                )
            )
            raise type(e)(
                "{!s}: Cannot set sweep stop of channel {!s} to {!s}\n{!s}".format(
                    __name__, channel, val, e
                )
            )
        return

    def _get_sweep_stop(self, channel=1):
        """
        Gets sweep stop value of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        val: float
            Stop value of sweep.
        """
        try:
            logging.debug(
                "{!s}: Get sweep stop of channel {:d}".format(__name__, channel)
            )
            return float(self.query(":sour{:d}:swe:stop".format(channel)))
        except Exception as e:
            logging.error(
                "{!s}: Cannot get sweep stop of channel {!s}".format(__name__, channel)
            )
            raise type(e)(
                "{!s}: Cannot get sweep stop of channel {!s}\n{!s}".format(
                    __name__, channel, e
                )
            )

    def _set_sweep_step(self, val, channel=1):
        """
        Sets sweep step value of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Step value of sweep.
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
                None
        """
        try:
            logging.debug(
                "{!s}: Set sweep step of channel {:d} to {:f}".format(
                    __name__, channel, val
                )
            )
            self.write("sour{:d}:swe:step {:f}".format(channel, val))
        except Exception as e:
            logging.error(
                "{!s}: Cannot set sweep step of channel {!s} to {!s}".format(
                    __name__, channel, val
                )
            )
            raise type(e)(
                "{!s}: Cannot set sweep step of channel {!s} to {!s}\n{!s}".format(
                    __name__, channel, val, e
                )
            )
        return

    def _get_sweep_step(self, channel=1):
        """
        Gets sweep step value of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        val:
            Step value of sweep.
        """
        try:
            logging.debug(
                "{!s}: Get sweep step of channel {:d}".format(__name__, channel)
            )
            return float(self.query("sour{:d}:swe:step?".format(channel)))
        except Exception as e:
            logging.error(
                "{!s}: Cannot get sweep step of channel {!s}".format(__name__, channel)
            )
            raise type(e)(
                "{!s}: Cannot get sweep step of channel {!s}\n{!s}".format(
                    __name__, channel, e
                )
            )

    def _get_sweep_nop(self, channel=1):
        """
        Gets sweeps number of points (nop) of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        val:
            Number of points of sweep.
        """
        try:
            logging.debug(
                "{!s}: Get sweep nop of channel {:d}".format(__name__, channel)
            )
            return int(
                (
                    self._get_sweep_stop(channel=channel)
                    - self._get_sweep_start(channel=channel)
                )
                / self._get_sweep_step(channel=channel)
                + 1
            )
        except Exception as e:
            logging.error(
                "{!s}: Cannot get sweep nop of channel {!s}".format(__name__, channel)
            )
            raise type(e)(
                "{!s}: Cannot get sweep nop of channel {!s}\n{!s}".format(
                    __name__, channel, e
                )
            )

    def set_sweep_mode(self, mode=1, *channels):
        """
        Setterfunction for the internal variable _sweep_mode.

        Sets an variable to indicate that
         * voltage is applied and current is measured (VI-mode)
        and channels <channels> that are used during sweeps.

        This function is necessary for the transport mode, however this device only
        supports this one mode.

        Parameters
        ----------
        mode: int. {1}
            Sweep mode denoting bias and sense modes. Possible values are 1 (VI-mode).
            Default is 1.
        channels: int or tuple(int).
            Number of channel of usage. Must be between 1 and 24 or a tuple containing
            two channels (1st argument as bias channel and 2nd argument as sense channel).


        Returns
        -------
        None
        """
        default_channels = (1, 1)
        if mode == 2:
            if len(channels) == 2:
                logging.debug(
                    "{!s}: Set sweep channels to {!s}".format(__name__, channels)
                )
                self._sweep_channels = channels
            elif len(channels) == 1:
                logging.debug(
                    "{!s}: Set sweep channels to {!s}".format(__name__, channels)
                )
                self._sweep_channels = (channels, channels)
            elif len(channels) == 0:
                logging.debug(
                    "{!s}: Set sweep channels to {!s}".format(
                        __name__, default_channels
                    )
                )
                self._sweep_channels = default_channels
            else:
                logging.error(
                    "{!s}: Cannot set sweep channels to {!s}".format(__name__, channels)
                )
                raise ValueError(
                    "{!s}: Cannot set sweep channels to {!s}".format(__name__, channels)
                )
            logging.debug("{!s}: Set sweep mode to {:d}".format(__name__, mode))
            self._sweep_mode = mode
        else:
            logging.error(
                "{!s}: Cannot set sweep mode to {!s}. Only mode 2 (VI-mode) is supported.".format(
                    __name__, mode
                )
            )
            raise ValueError(
                "{!s}: Cannot set sweep mode to {!s}".format(__name__, mode)
            )
        return

    def get_sweep_mode(self):
        """
        Getterfunction for internal variable _sweep_mode.

        Gets an internal variable <mode> to decide weather
         * 0: voltage is both applied and measured (VV-mode),
         * 1: current is applied and voltage is measured (IV-mode) or
         * 2: voltage is applied and current is measured (VI-mode).

        This device only supports mode 2 (VI-mode).

        Parameters
        ----------
        None

        Returns
        -------
        mode: int
            Sweep mode denoting bias and sense modes. Meanings are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode).
        """
        return self._sweep_mode

    def get_sweep_channels(self):
        """
        Gets channels <channels> that are used during sweeps.
        In the IV sweep mode one channel requires for biasing and sensing, whereas two channels are returned in the VV sweep mode (1st as bias and 2nd as sense channel).

        Parameters
        ----------
        None

        Returns
        -------
        channels: int or tuple(int)
            Number of channel of usage. Meanings are 1 or 2 for IV-mode and VI-mode or a tuple containing both channels for VV-mode (1st argument as bias channel and 2nd argument as sense channel).
        """
        return self._sweep_channels

    def get_sweep_bias(self):
        """
        Calls get_bias_mode of channel <channel>.
        This method is needed for qkit.measure.transport.transport.py in case of no virtual tunnel electronic.

        Parameters
        ----------
        None

        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        return self.get_bias_mode(self._sweep_channels[0])

    def set_status(self, status, channel=1):
        """
        Sets output DC status of channel <channel> to <status>.

        Parameters
        ----------
        status: int
            Output status. Possible values are 0 (off) or 1 (on).
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
                None
        """
        try:
            logging.debug(
                "{!s}: Set output status of channel {:d} to {!r}".format(
                    __name__, channel, status
                )
            )
            if status == 1:
                self._write(":sour{:d}:dc:init".format(channel))
            elif status == 0:
                self._write("sour{:d}:all:abor".format(channel))
                self._write("sour{:d}:volt 0".format(channel))
            else:
                raise TypeError("Status not valid, only 0 (off) and 1 (on) possible.")
        except Exception as e:
            logging.error(
                "{!s}: Cannot set output status of channel {!s} to {!s}".format(
                    __name__, channel, status
                )
            )
            raise type(e)(
                "{!s}: Cannot set output status of channel {!s} to {!s}\n{!s}".format(
                    __name__, channel, status, e
                )
            )
        return

    def get_end_of_sweep(self, channel=1):
        """
        Gets event of bias condition code register (ccr) entry "End for Sweep" <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be between 1 and 24. Default is 1.

        Returns
        -------
        val: bool
            End of sweep entry of bias condition code register.
        """
        try:
            logging.debug(
                """{!s}: Get bias ccr event "End of Sweep" of channel {:d}""".format(
                    __name__, channel
                )
            )
            return bool(
                (int(self._ask(":stat:sour:even")) >> (0 + 8 * (channel - 1))) % 2
            )
        except Exception as e:
            logging.error(
                '{!s}: Cannot get bias ccr event "End of Sweep" of channel {!s}'.format(
                    __name__, channel
                )
            )
            raise type(e)(
                '{!s}: Cannot get bias ccr event "End of Sweep" of channel {!s}\n{!s}'.format(
                    __name__, channel, e
                )
            )

    def _wait_for_end_of_sweep(self, channel=1):
        """
        Waits until the event of bias condition code register (ccr) entry "End for Sweep" of channel <channel> occurs.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        None
        """
        while not (self.get_end_of_sweep(channel=channel)):
            time.sleep(100e-3)
        return

    def get_end_of_measure(self, channel=1):
        """
        Gets condition of sense condition code register (ccr) entry "End of Measure" <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: bool
            End of sweep entry of bias condition code register.
        """
        # Corresponding Command: :STATus:SENSe:CONDition?
        try:
            logging.debug(
                """{!s}: Get sense ccr event "End of Measure" of channel {:d}""".format(
                    __name__, channel
                )
            )
            return bool(
                (int(self._ask(":stat:sens:cond")) >> (0 + 8 * (channel - 1))) % 2
            )
        except Exception as e:
            logging.error(
                '{!s}: Cannot get sense ccr event "End of Measure" of channel {!s}'.format(
                    __name__, channel
                )
            )
            raise type(e)(
                '{!s}: Cannot get sense ccr event "End of Measure" of channel {!s}\n{!s}'.format(
                    __name__, channel, e
                )
            )



"""
This module provides drivers for accessing the IVVI's Bias DACs. 

For more information about the IVVI visit https://qtwork.tudelft.nl/~schouten/ivvi/index-ivvi.htm .
The underlying communication protocol is explained at https://qtwork.tudelft.nl/~schouten/ivvi/doc-d5/rs232linkformat.txt .

Improvements compared to legacy codebase include
- added various serial port connection types in SerialPortCommunication module
- underlying driver version readback
- internal book-keeping of set DAC-range-knobs using helper class IVVI_DAC_Group to allow adjusted conversions at different DACs
- enhanced qkit conformity and integration
- dummy IVD for sweeping given ranges in different modes

Author: Marius Frohn <uzrfo@student.kit.edu>
Version: 2.4; (08/2024) 

(Legacy DOC-string)
IVVI.py class, to perform the communication between the Wrapper and the device
Pieter de Groot <pieterdegroot@gmail.com>, 2008
Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
Reinier Heeres <reinier@heeres.eu>, 2008

Extended by Tim Wolz to access the IVVI via Ethernet.
Data is sent as a string via Ethernet to a Raspberry Pi, 
where it is converted to asci code and sent to the IVVI
Ethernet connection based on the lazy pirate pattern by Daniel Lundin <dln(at)eintr(dot)org> 
(Note Marius: Moved to SerialPortCommunication.SerialPortRasbpi)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import logging, time
from qkit.core.instrument_base import Instrument
import SerialPortCommunication as SPC
import numpy as np

class IVVI_BiasDAC(Instrument):
    """
    This driver provides access to the IVVI's bias DACs.
    Usage requires communicating with the IVVI's control box via RS232 serial port connection.
    The SerialPortCommunication module/instrument is intended to do that.

    Initialize with
    <SPC_name> = instruments.create('<SPC_name>', 'SerialPortCommunication', *args for creating connection*)
    <name> = instruments.create('<name>', 'IVVI_BiasDAC', connection=<SPC_name>, **kwargs)
    """

    class IVVI_DAC_Group:
        """
        Helper class for keeping track of set DAC ranges at IVVI D5 module and binary <-> V conversion
        IVVI can switch DAC polization between negative, centered around 0 and positive in groups of 4. 
        """
        def __init__(self, val1: float, val2: float):
            self.low = min(val1, val2)
            self.high = max(val1, val2)
        def __str__(self) -> str:
            return "IVVI DAC range helper object for {} to {} V".format(self.low, self.high)
        def __hash__(self):
            return hash(tuple((self.low, self.high)))
        def v2bytes(self, volt: float) -> tuple[int, int]:
            if volt < self.low or volt > self.high:
                raise ValueError
            else:
                bytevalue = int(round(65535*(volt - self.low)/(self.high - self.low)))
                return (bytevalue//256, bytevalue % 256)
        def bytes2v(self, byte_high: int, byte_low: int) -> float:
            # Correct readback expected, no value checking
            return self.low + (self.high - self.low)*(256*byte_high + byte_low)/65536

    # available ranges
    DACRangeNeg = IVVI_DAC_Group(-4.0, 0.0)
    DACRangeBip = IVVI_DAC_Group(-2.0, 2.0)
    DACRangePos = IVVI_DAC_Group(0.0, 4.0)
    RangeIdent = {  "pos": DACRangePos, "bip": DACRangeBip, "neg": DACRangeNeg, 
                    "Pos": DACRangePos, "Bip": DACRangeBip, "Neg": DACRangeNeg, 
                    "POS": DACRangePos, "BIP": DACRangeBip, "NEG": DACRangeNeg, 
                    "+"  : DACRangePos, "0"  : DACRangeBip, "-"  : DACRangeNeg  }
    ReverseIdent = { DACRangePos: "Pos", DACRangeBip: "Bip", DACRangeNeg: "Neg" }
    
    class DummyIVD:
        """
        This is a primitive dummy for using the IVVI as a IVD. 
        It only takes the value-setting side into consideration (voltage divider v_div or current source dAdV), 
        while e.g. amplification factors when measuring values should be handled within the measure-function.
        Setting status on/off of additionally needed devices should also be handled outside of this class
        """
        def __init__(self):
            """
            These parameters need to be set before sweeping
            """
            self.measure_func = None
            self.sweep_channel = None
            self.delay_setget = None
            self.pseudo_bias_mode = None # 0: current bias; 1: voltage bias
            self.dAdV = None # for current bias
            self.v_div = None # for voltage bias
        def get_sweep_mode(self):
            # 0: V bias V measure, 1: I bias V measure, 2: V bias I measure
            # technically mode 2 also possible, this class cant determine 0 vs 2 without hidden knowledge about measure function though
            return 1 - self.pseudo_bias_mode
        def get_sweep_bias(self):
            return self.pseudo_bias_mode
        def get_sweep_channels(self):
            return [self.sweep_channel]
        def set_status(self, *args, **kwargs):
            pass

    def __init__(self, name: str, connection: SPC.SerialPortCommunication, knob_config: list[str] = ["0", "0", "0", "0"], reset_on_init: bool = True, make_dummy_ivd: bool = False):
        """
        Initialzes the IVVI Bias DAC by communicating via serial port

        Input:
            name (str)                              : name of the instrument
            connection (SerialPortCommunication)    : instrument handling communication, see module
            knob_config (list[str])                 : initially set DAC group ranges
            reset_on_init (bool)                    : reset all dacs to 0 after initialization or keep as is
            make_dummy_ivd (bool)                   : include parameters for using IVVI as IVD

        Output:
            None
        """
        self.dummy_ivd = self.DummyIVD() if make_dummy_ivd else None
        Instrument.__init__(self, name, tags=['physical'])
        self.connection = connection
        for j in range(16):
            self.__dict__["do_set_DAC_{}_voltage".format(j + 1)] = lambda mVolt, j=j: self.set_dac(j + 1, mVolt)
            self.__dict__["do_get_DAC_{}_voltage".format(j + 1)] = lambda j=j: self.get_dac(j + 1)
            self.add_parameter("DAC_{}_voltage".format(j + 1), type=float, minval=self.DACRangeBip.low, maxval=self.DACRangeBip.high, units='V', flags=Instrument.FLAG_GETSET)
        self.dac_ranges = [self.DACRangeBip] * 4
        for i in range(4):
            self.set_dacgroup_range(i, knob_config[i])
        self.add_function("set_dacgroup_range")
        self.add_function("reset_dacs")
        self.add_function("get_driver_version")
        if reset_on_init:
            self.reset_dacs()
        logging.info("{} : Initializing instrument IVVI via serial port with driver version {}".format(self._name, self.get_driver_version()))
    
    def __del__(self):
        """
        Closes up the IVVI driver

        Input:
            None

        Output:
            None
        """
        del self.connection
        logging.info(__name__ + ': Deleting IVVI instrument')
    
    def __getattr__(self, name):
        """
        Add DummyIVD attributes to class, if enabled
        """
        if not self.dummy_ivd is None:
            return self.dummy_ivd.__getattribute__(name)
        else:
            print("Cannot find attribute '{}' for IVVI_BiasCAC, perhaps dummy IVD needs to be enabled".format(name))
            raise AttributeError
    
    @staticmethod
    def handle_readback_error_flag(readback: list[int]) -> list[int]:
        """
        Readback of every command sent to IVVI-rack has the format [*Readback length*, *Error flag byte*, DataByte, DataByte, ...].
        This function checks if readback has in fact the size it claims to have and interprets the error flag.
        It strips these two bytes from the readback and returns only the data bytes.

        Input:
            readback (list[int]): Readback from a communication with IVVI-rack

        Output:
            (list[int]): Data bytes of given readback
        """
        if len(readback) != readback[0]:
            logging.info(__name__ + ": Unknown communication error")
            print(__name__ + ": Unknown communication error")
        if readback[1] == 32:
            logging.info(__name__ + ': Watchdog reset detected')
            print(__name__ + ": Watchdog reset detected")
        elif readback[1] == 64:
            logging.info(__name__ + ': DAC does not exist')
            print(__name__ + ": DAC does not exist")
        elif readback[1] == 128:
            logging.info(__name__ + ': wrong action')
            print(__name__ + ": wrong action")
        return readback[2:]
      
    def set_dacgroup_range(self, group_num: int, ident: str):
        """
        Does NOT change actual polization of the DACs, only possible via knobs at IVVI. 
        Set internal values to according analog positions for correct value conversions!

        Input: 
            group_num (int) : 0 ... 3; index of the DAC group corresponding to DACs 1-4, 5-8, ...
            ident (str)     : Identifier for DAC range. See IVVIDIG_main_eth.RangeIdent for possibilities

        Output:
            None
        """
        if not ident in self.RangeIdent:
            print("Given identifier '{}' unkown. Try 'pos', 'bip', 'neg' instead. Group range unchanged".format(ident))
        elif group_num < 0 or group_num >= len(self.dac_ranges):
            print("DAC group number {} out of range 0 .. 3. Group range unchanged".format(group_num))
        else:
            self.dac_ranges[group_num] = self.RangeIdent[ident]
            for i in range(4):
                self.set_parameter_options("DAC_{}_voltage".format(4*group_num + i + 1), minval=self.dac_ranges[group_num].low, maxval=self.dac_ranges[group_num].high)
    
    def reset_dacs(self):
        """
        Set all DACs to 0 V, adjusted to their currently set range

        Input:
            None

        Output:
            None
        """
        for channel in range(16):
            self.set_dac(channel + 1, 0)
    
    def set_dac(self, channel: int, volt: float):
        """
        Set DAC channel to given voltage, if value is valid within set range

        Input:
            channel (int) : 1-based channel index (DAC #1 -> channel = 1, not 0)
            volt (float) : to be set voltage in V
        
        Output:
            None
        """
        try:
            self.handle_readback_error_flag(self.connection.communicate([7, 0, 2, 1, channel] + list(self.dac_ranges[(channel - 1)//4].v2bytes(volt)), 2))
        except ValueError:
            print("Given voltage {} V not within selected DAC-range {} .. {} V".format(volt, self.dac_ranges[(channel - 1)//4].low, self.dac_ranges[(channel - 1)//4].high))

    def get_dacs_raw(self) -> list[int]:
        """
        Read back all DACs and return data bytes

        Input:
            None

        Output:
            (list[int]) : raw DAC values as [DAC_1_byte_high, DAC_1_byte_low, DAC_2_byte_high, ...]
        """
        return self.handle_readback_error_flag(self.connection.communicate([4, 0, 34, 2], 34))
    
    def get_dacs_list(self) -> list[float]:
        """
        Returns list of set DAC voltages in V

        Input:
            None

        Output:
            (list[float]) : set DAC voltages in V
        """
        data = self.get_dacs_raw()
        return [self.dac_ranges[channel//4].bytes2v(data[2*channel], data[2*channel + 1]) for channel in range(16)]

    def get_dac(self, channel: int) -> float:
        """
        Returns set voltage at specific DAC channel

        Input: 
            channel (int) : 1-based channel index (DAC #1 -> channel = 1, not 0)

        Output:
            (float) : voltage in V
        """
        data = self.get_dacs_raw()
        return self.dac_ranges[(channel - 1)//4].bytes2v(data[2*channel - 2], data[2*channel - 1]) 

    def get_driver_version(self) -> int:
        """
        Returns version number of underlying driver

        Input:
            None

        Output:
            (int) : version number
        """
        # For some unknown reason reading out the driver version outmatically sets it as high byte in DAC1
        # This call fixes this error in an ugly way
        dac_val = self.get_dac(1)
        drive = self.handle_readback_error_flag(self.connection.communicate([4, 0, 3, 4], 3))[0]
        self.set_dac(1, dac_val)
        return drive
    
    def take_IV(self, sweep):
        if self.dummy_ivd is None:
            print("IVVI_BiasDAC needs dummy IVD enabled to take IV-curve")
            raise AttributeError
        start, stop, step, _ = sweep
        setty = []
        getty = []
        for val in np.linspace(start, stop, int(round(np.abs(start-stop)/step+1))):
            if self.dummy_ivd.pseudo_bias_mode: # voltage bias
                self.set_dac(self.dummy_ivd.sweep_channel, val*self.dummy_ivd.v_div)
                setty += [self.get_dac(self.dummy_ivd.sweep_channel)/self.dummy_ivd.v_div]
            else: # current bias
                self.set_dac(self.dummy_ivd.sweep_channel, val/self.dummy_ivd.dAdV)
                setty += [self.get_dac(self.dummy_ivd.sweep_channel)*self.dummy_ivd.dAdV]
            time.sleep(self.dummy_ivd.delay_setget)
            getty += [self.dummy_ivd.measure_func()]
        # return requires I-values first, V-values second independent of what was set/measured
        if self.dummy_ivd.pseudo_bias_mode: # voltage bias
            return np.array(getty), np.array(setty)
        else: # current bias
            return np.array(setty), np.array(getty)

    # qkit parameter documentation
    def get_parameters(self) -> dict:
        """
        Returns info for qkit
        """
        # DAC values in V
        params = {"DAC_{}_voltage".format(i+1): {} for i in range(16)}
        if self.dummy_ivd is None:
            ivd_params = {}
        else:
            ivd_params = {
                "measure_func": {},
                "sweep_channel": {},
                "delay_setget": {},
                "pseudo_bias_mode": {}, 
                "dAdV": {},
                "v_div": {}
            }
        return {**params, **ivd_params}
    
    def get_measure_func(self):
            return self.dummy_ivd.measure_func.__name__ if self.dummy_ivd is not None else None
    def get_sweep_channel(self):
        return self.dummy_ivd.sweep_channel if self.dummy_ivd is not None else None
    def get_delay_setget(self):
        return self.dummy_ivd.delay_setget if self.dummy_ivd is not None else None
    def get_pseudo_bias_mode(self):
        return self.dummy_ivd.pseudo_bias_mode if self.dummy_ivd is not None else None
    def get_dAdV(self):
        return self.dummy_ivd.dAdV if self.dummy_ivd is not None else None
    def get_v_div(self):
        return self.dummy_ivd.v_div if self.dummy_ivd is not None else None
    def get(self, param, **kwargs):
        try:
            return eval("self.get_{}()".format(param))
        except:
            return None



import qkit
import os
import logging
from time import strftime


def cleanup_logfiles():
    '''
    if qkit.cfg['maintain_logiles'] is not False, this script checks the log folder and removes all log files except for the latest 10.
    :return:
    '''
    if qkit.cfg.get('maintain_logfiles',True):
        ld = [filename for filename in os.listdir(qkit.cfg['logdir']) if filename.startswith('qkit') and filename.endswith('.log')]
        ld.sort()
        for f in ld[:-10]:
            try:
                os.remove(os.path.join(qkit.cfg['logdir'], f))
            except:
                pass


def _setup_logging():
    fileLogLevel = getattr(logging, qkit.cfg.get('file_log_level', 'WARNING'))
    stdoutLogLevel = getattr(logging, qkit.cfg.get('stdout_log_level', 'WARNING'))
    
    rootLogger = logging.getLogger()
    
    fileLogger = logging.FileHandler(filename=os.path.join(qkit.cfg['logdir'], strftime('qkit_%Y%m%d_%H%M%S.log')), mode='a+')
    fileLogger.setFormatter(
            logging.Formatter('%(asctime)s %(levelname)-8s: %(message)s (%(filename)s:%(lineno)d)',
                              datefmt='%Y-%m-%d %H:%M:%S'))
    fileLogger.setLevel(fileLogLevel)
    
    jupyterLogger = logging.StreamHandler()
    jupyterLogger.setFormatter(
            logging.Formatter('%(asctime)s [%(levelname)-8s]: %(message)s (%(filename)s:%(lineno)d)',
                              datefmt='%Y-%m-%d %H:%M:%S'))
    jupyterLogger.setLevel(stdoutLogLevel)
    
    rootLogger.addHandler(fileLogger)
    rootLogger.addHandler(jupyterLogger)
    rootLogger.setLevel(min(stdoutLogLevel, fileLogLevel))
    
    logging.info(' ---------- LOGGING STARTED ---------- ')
    
    logging.debug('Set logging level for file to: %s ' % fileLogLevel)
    logging.debug('Set logging level for stdout to: %s ' % stdoutLogLevel)
    
    cleanup_logfiles()

def set_debug(enable):
    logger = logging.getLogger()
    if enable:
        logger.setLevel(logging.DEBUG)
        logging.info('Set logging level to DEBUG')
    else:
        logger.setLevel(logging.INFO)
        logging.info('Set logging level to INFO')


_setup_logging()



import qkit

import os
import threading
import logging
import time
import json
import numpy as np
import qkit.storage.hdf_DateTimeGenerator as dtg
from qkit.core.lib.file_service.breadcrumbs import BreadCrumbCreator
import h5py

try:
    import cPickle as pickle
except:
    import pickle

class UUID_base(object):
    _alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    def get_uuid(self,time):
        """
        
        returns a UUID from a given time, e.g. returned by time.time()
        The UUID is returned with a precision of (integer) seconds 
        and has a fixed length of six characters.
        
        Derived from encode_uuid(), orginally located in hdf_DateTimeGenerator.py (AS/MP/HR)
        """
        # if not value: value = self._unix_timestamp
        output = ''
        time = int(time)
        la = len(self._alphabet)
        while time:
            output += self._alphabet[time % la]
            time = int(time / la)
        return output[::-1]
    
    def get_time(self,uuid):
        """
        returns a integer time value from a given UUID timestamp (reverse of get_UUID())
        orginally located in hdf_DateTimeGenerator.py (AS/MP/HR)
        """
        # if not string: string = self._uuid
        output = 0
        multiplier = 1
        uuid = uuid[::-1].upper()
        la = len(self._alphabet)
        while uuid != '':
            f = self._alphabet.find(uuid[0])
            if f == -1:
                raise ValueError("fid.get_time: Can not decode this: {}<--".format(uuid[::-1]))
            output += f * multiplier
            multiplier *= la
            uuid = uuid[1:]
        return output
        
    def get_date(self,uuid):
        """
        Returns a date string from a given UUID timestamp (reverse of get_uuid())
        """
        return time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(self.get_time(uuid)))


class file_system_service(UUID_base):
    h5_db = {}
    set_db = {}
    measure_db = {}
    h5_info_db = {}

    _h5_n_mtime = {}
    
    _h5_mtime_db_path   = os.path.join(qkit.cfg['logdir'],"h5_mtime.db")
    _h5_info_cache_path = os.path.join(qkit.cfg['logdir'],"h5_info_cache.db")

    _breadcrumb_creator = BreadCrumbCreator()

    lock = threading.Lock()
    
    def _remove_cache_files(self):
        """
            remove cached files to recreate the database
        """
        for f in [self._h5_mtime_db_path,self._h5_info_cache_path]:
            if os.path.isfile(f):
                os.remove(f)

    def _load_cache_files(self):
        """ to speed up things, try to load the h5 
            information from previous runs. 
        """
        self._new_cache = False
        self._h5_mtime_db = {}
        self._h5_info_cache_db = {}
        self._n_mtime = {}
        try:
            with open(self._h5_mtime_db_path,'rb') as f:
                self._h5_mtime_db = pickle.load(f)
            with open(self._h5_info_cache_path,'rb') as f:
                self._h5_info_cache_db = pickle.load(f)
        except IOError as e:
            logging.info("m_time_db not found. Not using cached files for now. %s"%e)
            self._new_cache = True

    def _store_cache_files(self):
        """ writes the cached files on disk, 
        qkit.cfg['logdir'] is used by default.

        exeptions are not handled ... if something fails here 
        there is a severe general problem, e.g. the disk is full.
        """
        write_protocol = 2 # For Python 2 compatibility 0=text, 1...x binary, -1 highest binary. 
        with open(self._h5_mtime_db_path,'wb+') as f:
            pickle.dump(self._h5_n_mtime,f,protocol=write_protocol)
        with open(self._h5_info_cache_path,'wb+') as f:
            pickle.dump(self.h5_info_db,f,protocol=write_protocol)

    def _get_datadir(self):
        if qkit.cfg.get('fid_restrict_to_userdir',False):
            return os.path.split(dtg.DateTimeGenerator().new_filename()['_folder'])[0]
        else:
            return qkit.cfg['datadir']

    def update_file_db(self):
        with self.lock:
            start_time = time.time()
            self._load_cache_files()
            if qkit.cfg.get('fid_scan_datadir',True):
                qkit.cfg['fid_scan_datadir'] = True
                logging.debug("file info database: Start to update database.")
                for root, _ , files in os.walk(self._get_datadir()): # root, dirs, files
                    for f in files:
                        self._inspect_and_add_Leaf(f,root)
                logging.debug("file info database: Updating database done.")
            self._store_cache_files()
            print ("Initialized the file info database (qkit.fid) in %.3f seconds."%(time.time()-start_time))

    def _inspect_and_add_Leaf(self,fname,root):
        """
        inspect the filenames if .h5, .set or .measurement

        to speed up things, the files are only scanned 
        if something has changed (os.stat.m_time) on disk. 
        """

        # join to absolute path:
        fqpath = os.path.join(root, fname)

        # take the prefix ...
        uuid = fname[:6]

        # ... and check the suffix
        if fqpath[-3:] == '.h5':

            # save the path using uuids as an index
            # Note: All path entries with the same uuid are 
            # overwritten with the last found uuid indexed file
            self.h5_db[uuid] = fqpath
            self._breadcrumb_creator.append_entry(uuid, fqpath)

            # we only care about the mtime of .h5 files 
            mtime = os.stat(fqpath).st_mtime

            # store the file's modification time 
            self._h5_n_mtime[uuid] = mtime

            if mtime == self._h5_mtime_db.get(uuid,0):
                if self._h5_info_cache_db.get(uuid,0):
                    self.h5_info_db[uuid] = self._h5_info_cache_db.get(uuid)
                else:
                    self._collect_info(uuid, fqpath) # collect_info is expensive.
            else:
                self._collect_info(uuid, fqpath) # collect_info is expensive. 

        elif fqpath[-3:] == 'set':
            self.set_db[uuid] = fqpath
        elif fqpath[-3:] == 'ent':
            self.measure_db[uuid] = fqpath

    def _collect_info(self,uuid,path):
            tm = ""
            dt = ""
            j_split = (path.replace('/', '\\')).split('\\')
            name = j_split[-1][7:-3]
            if ord(uuid[0]) > ord('L'):
                try:
                    tm = self.get_time(uuid)
                    dt = self.get_date(uuid)
                    user = j_split[-3]
                    run = j_split[-4]
                except ValueError as e:
                    user = None
                    run = None
                    logging.info(e)
            else:
                tm = uuid
                try:
                    if j_split[-3][0:3] != 201:  # not really a measurement file then
                        dt = None
                    else:
                        dt = '{}-{}-{} {}:{}:{}'.format(j_split[-3][:4], j_split[-3][4:6], j_split[-3][6:], tm[:2], tm[2:4], tm[4:])
                except IndexError:
                    dt = None
                user = None
                run = None
            h5_info_db = {'time': tm, 'datetime': dt, 'run': run, 'name': name, 'user': user}
            
            if qkit.cfg.get('fid_scan_hdf', False):
                h5_info_db.update({'rating':10})
                try:
                    h5f=h5py.File(path,'r')
                    if "comment" in  h5f['/entry/data0'].attrs:
                        h5_info_db.update({'comment': h5f['/entry/data0'].attrs['comment']})
                    if "dr_values" in h5f['/entry/analysis0']:
                        try:
                            # this is legacy and should be removed at some point
                            # please use the entry/analysis0 attributes instead.
                            fit_comment = h5f['/entry/analysis0/dr_values'].attrs.get('comment',"").split(', ')
                            comm_begin = [i[0] for i in fit_comment]
                            try:
                                h5_info_db.update({'fit_freq': float(h5f['/entry/analysis0/dr_values'][comm_begin.index('f')])})
                            except (ValueError, IndexError):
                                pass
                            try:
                                h5_info_db.update({'fit_time': float(h5f['/entry/analysis0/dr_values'][comm_begin.index('T')])})
                            except (ValueError, IndexError):
                                pass
                        except (KeyError, AttributeError):
                            pass
                    if "measurement" in h5f['/entry/data0']:
                        try:
                            mmt = json.loads(h5f['/entry/data0/measurement'][0])
                            h5_info_db.update(
                                    {arg: mmt[arg] for arg in ['run_id', 'user', 'rating', 'smt'] if mmt.has_key(arg)}
                            )
                        except(AttributeError, KeyError):
                            pass
                    try:
                        h5_info_db.update(dict(h5f['/entry/analysis0'].attrs))
                    except(AttributeError, KeyError):
                        pass
                except KeyError as e:
                    logging.debug("fid could not index file {}, probably it is just new and empty. Original message: {}".format(path,e))
                except IOError as e:
                    logging.error("fid {}:{}".format(path,e))
                finally:
                    h5f.close()

            self.h5_info_db[uuid] = h5_info_db
    
    def add_h5_file(self, h5_filename):
        if qkit.cfg['fid_scan_datadir']:
            threading.Timer(20, function=self._add, kwargs={'h5_filename':h5_filename}).start()
        
    def _add(self, h5_filename):
        """
        Add a file to the database, where h5_filename should be the absolute filepath to the h5 file.
        """
        basename = os.path.basename(h5_filename)[:-2]
        dirname = os.path.dirname(h5_filename)
        uuid = basename[:6]
        self._breadcrumb_creator.append_entry(uuid, h5_filename)
        if h5_filename[-3:] != '.h5':
            logging.error("Tried to add '{:s}' to the qkit.fid database: Not a .h5 filename.".format(h5_filename))
        with self.lock:
            if os.path.isfile(h5_filename):
                logging.debug("Store_db: Adding manually h5: " + basename + 'h5')
                self._inspect_and_add_Leaf(basename + 'h5', dirname)
            else:
                logging.error("Tried to add '{:s}' to the qkit.fid database: File does not exist.".format(h5_filename))
            if os.path.isfile(h5_filename[:-2] + 'set'):
                logging.debug("Store_db: Adding manually set: " + basename + 'set')
                self._inspect_and_add_Leaf(basename + 'set', dirname)
            if os.path.isfile(h5_filename[:-2] + 'measurement'):
                logging.debug("Store_db: Adding manually measurement: " + basename + 'measurement')
                self._inspect_and_add_Leaf(basename + 'measurement', dirname)
        self.update_grid_db()


    def _set_hdf_attribute(self,UUID,attribute,value):
        h5_filepath = self.h5_db[UUID]
        h = h5py.File(h5_filepath,'r+')['entry']
        try:
            if not 'analysis0' in h:
                h.create_group('analysis0')
            if value=="":
                if attribute in h['analysis0'].attrs:
                    del h['analysis0'].attrs[attribute]
            else:
                if qkit.module_available['pandas']:
                    import pandas as pd
                    if pd.isnull(value):
                        if attribute in h['analysis0'].attrs:
                            del h['analysis0'].attrs[attribute]
                    else:
                        h['analysis0'].attrs[attribute] = value
                else:
                    h['analysis0'].attrs[attribute] = value
        finally:
            h.file.close()
        self.h5_info_db[UUID].update({attribute:value})
        
    def wait(self):
        with self.lock:
            pass
        qkit.flow.sleep(.1) #to prevent timing issues
        if self.lock.locked():
            self.wait()
        return True


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
@author: dennis.rieger@kit.edu / 2019

inspired by and based on resonator tools of Sebastian Probst
https://github.com/sebastianprobst/resonator_tools
"""

import numpy as np
import logging
import scipy.optimize as spopt
from scipy import stats
from scipy.interpolate import splrep, splev
from scipy.ndimage.filters import gaussian_filter1d
plot_enable = False
try:
    import qkit
    if qkit.module_available("matplotlib"):
        import matplotlib.pyplot as plt
        plot_enable = True
except (ImportError, AttributeError):
    try:
        import matplotlib.pyplot as plt
        plot_enable = True
    except ImportError:
        plot_enable = False

class circuit:
    """
    Base class for common routines and definitions shared between both ports.
    
    inputs:
    - f_data: Frequencies for which scattering data z_data_raw is taken
    - z_data_raw: Measured values for scattering parameter S11 or S21 taken at
                  frequencies f_data
    """
    
    def __init__(self, f_data, z_data_raw=None):
        self.f_data = np.array(f_data)
        self.z_data_raw = np.array(z_data_raw)
        self.z_data_norm = None
        
        self.fitresults = {}
        
        self.fit_delay_max_iterations = 5
    
    @classmethod
    def Sij(cls, f, fr, Ql, Qc, phi=0., a=1., alpha=0., delay=0.):
        """
        Full model for S11 of single-port reflection measurements or S21 of 
        notch configuration measurements. The models only differ in a factor of
        2 which is set by the dedicated port classes inheriting from this class.
        
        inputs:
        - fr: Resonance frequency
        - Ql: Loaded quality factor
        - Qc: Coupling (aka external) quality factor. Calculated with
              diameter correction method, i.e. 1/Qc = Re{1/|Qc| * exp(i*phi)}
        - phi (opt.): Angle of the circle's rotation around the off-resonant
                      point due to impedance mismatches or Fano interference.
        - a, alpha (opt.): Arbitrary scaling and rotation of the circle w.r.t. 
                           origin
        - delay (opt.): Time delay between output and input signal leading to
                        linearly frequency dependent phase shift
        """
        complexQc = Qc*np.cos(phi)*np.exp(-1j*phi)
        return a*np.exp(1j*(alpha-2*np.pi*f*delay)) * (
            1. - 2.*Ql / (complexQc * cls.n_ports * (1. + 2j*Ql*(f/fr-1.)))
        )
    
    def autofit(self, calc_errors=True, fixed_delay=None, isolation=15):
        """
        Automatically calibrate data, normalize it and extract quality factors.
        If the autofit fails or the results look bad, please discuss with
        author.
        """
        
        if fixed_delay is None:
            self._fit_delay()
        else:
            self.delay = fixed_delay
            # Store result in dictionary (also for backwards-compatibility)
            self.fitresults["delay"] = self.delay
        self._calibrate()
        self._normalize()
        self._extract_Qs(calc_errors=calc_errors)
        self.calc_fano_range(isolation=isolation)
        
        # Prepare model data for plotting
        self.z_data_sim = self.Sij(
            self.f_data, self.fr, self.Ql, self.Qc, self.phi,
            self.a, self.alpha, self.delay
        )
        self.z_data_sim_norm = self.Sij(
            self.f_data, self.fr, self.Ql, self.Qc, self.phi
        )
    
    def _fit_delay(self):
        """
        Finds the cable delay by repeatedly centering the "circle" and fitting
        the slope of the phase response.
        """
        
        # Translate data to origin
        xc, yc, r0 = self._fit_circle(self.z_data_raw)
        z_data = self.z_data_raw - complex(xc, yc)
        # Find first estimate of parameters
        fr, Ql, theta, self.delay = self._fit_phase(z_data)
        
        # Do not overreact (see end of for loop)
        self.delay *= 0.05
        
        # Iterate to improve result for delay
        for i in range(self.fit_delay_max_iterations):
            # Translate new best fit data to origin
            z_data = self.z_data_raw * np.exp(2j*np.pi*self.delay*self.f_data)
            xc, yc, r0 = self._fit_circle(z_data)
            z_data -= complex(xc, yc)
            
            # Find correction to current delay
            guesses = (fr, Ql, 5e-11)
            fr, Ql, theta, delay_corr = self._fit_phase(z_data, guesses)
            
            # Stop if correction would be smaller than "measurable"
            phase_fit = self.phase_centered(self.f_data, fr, Ql, theta, delay_corr)
            residuals = np.unwrap(np.angle(z_data)) - phase_fit
            if 2*np.pi*(self.f_data[-1]-self.f_data[0])*delay_corr <= np.std(residuals):
                break
            
            # Avoid overcorrection that makes procedure switch between positive
            # and negative delays
            if delay_corr*self.delay < 0: # different sign -> be careful
                if abs(delay_corr) > abs(self.delay):
                    self.delay *= 0.5
                else:
                    # delay += 0.1*delay_corr
                    self.delay += 0.1*np.sign(delay_corr)*5e-11
            else: # same direction -> can converge faster
                if abs(delay_corr) >= 1e-8:
                    self.delay += min(delay_corr, self.delay)
                elif abs(delay_corr) >= 1e-9:
                    self.delay *= 1.1
                else:
                    self.delay += delay_corr
        
        if 2*np.pi*(self.f_data[-1]-self.f_data[0])*delay_corr > np.std(residuals):
            logging.warning(
                "Delay could not be fit properly!"
            )
        
        # Store result in dictionary (also for backwards-compatibility)
        self.fitresults["delay"] = self.delay
    
    def _calibrate(self):
        """
        Finds the parameters for normalization of the scattering data. See
        Sij for explanation of parameters.
        """
        
        # Correct for delay and translate circle to origin
        z_data = self.z_data_raw * np.exp(2j*np.pi*self.delay*self.f_data)
        xc, yc, self.r0 = self._fit_circle(z_data)
        zc = complex(xc, yc)
        z_data -= zc
        
        # Find off-resonant point by fitting offset phase
        # (centered circle corresponds to lossless resonator in reflection)
        self.fr, self.Ql, theta, self.delay_remaining = self._fit_phase(z_data)
        self.theta = self._periodic_boundary(theta)
        beta = self._periodic_boundary(theta - np.pi)
        offrespoint = zc + self.r0*np.cos(beta) + 1j*self.r0*np.sin(beta)
        self.offrespoint = offrespoint
        self.a = np.absolute(offrespoint)
        self.alpha = np.angle(offrespoint)
        self.phi = self._periodic_boundary(beta - self.alpha)
        
        # Store radius for later calculation
        self.r0 /= self.a
        
        # Store results in dictionary (also for backwards-compatibility)
        self.fitresults.update({
            "delay_remaining": self.delay_remaining,
            "a": self.a,
            "alpha": self.alpha,
            "theta": self.theta,
            "phi": self.phi,
            "fr": self.fr,
            "Ql": self.Ql
        })
    
    def _normalize(self):
        """
        Transforms scattering data into canonical position with off-resonant
        point at (1, 0) (does not correct for rotation phi of circle around
        off-resonant point).
        """
        self.z_data_norm = self.z_data_raw / self.a*np.exp(
            1j*(-self.alpha + 2.*np.pi*self.delay*self.f_data)
        )
        
    def _extract_Qs(self, refine_results=False, calc_errors=True):
        """
        Calculates Qc and Qi from radius of circle. All needed info is known
        already from the calibration procedure.
        """
        
        self.absQc = self.Ql / (self.n_ports*self.r0)
        # For Qc, take real part of 1/(complex Qc) (diameter correction method)
        self.Qc = self.absQc / np.cos(self.phi)
        self.Qi = 1. / (1./self.Ql - 1./self.Qc)
        self.Qi_no_dia_corr = 1. / (1./self.Ql - 1./self.absQc)
        
        # Store results in dictionary (also for backwards-compatibility)
        self.fitresults.update({
            "fr": self.fr,
            "Ql": self.Ql,
            "Qc": self.Qc,
            "Qc_no_dia_corr": self.absQc,
            "Qi": self.Qi,
            "Qi_no_dia_corr": self.Qi_no_dia_corr ,
        })
        
        # Calculate errors if wanted
        if calc_errors:
            chi_square, cov = self._get_covariance()
    
            if cov is not None:
                fr_err, Ql_err, absQc_err, phi_err = np.sqrt(np.diag(cov))
                # Calculate error of Qi with error propagation
                # without diameter correction
                dQl = 1. / ((1./self.Ql - 1./self.absQc) * self.Ql)**2
                dabsQc = -1. / ((1./self.Ql - 1./self.absQc) * self.absQc)**2
                Qi_no_dia_corr_err = np.sqrt(
                    dQl**2*cov[1][1]
                    + dabsQc**2*cov[2][2]
                    + 2.*dQl*dabsQc*cov[1][2]
                )
                # with diameter correction
                dQl = 1. / ((1./self.Ql - 1./self.Qc) * self.Ql)**2
                dabsQc = -np.cos(self.phi) / (
                    (1./self.Ql - 1./self.Qc) * self.absQc
                )**2
                dphi = -np.sin(self.phi) / (
                    (1./self.Ql - 1./self.Qc)**2 * self.absQc
                )
                Qi_err = np.sqrt(
                    dQl**2*cov[1][1]
                    + dabsQc**2*cov[2][2]
                    + dphi**2*cov[3][3]
                    + 2*(
                        dQl*dabsQc*cov[1][2]
                        + dQl*dphi*cov[1][3]
                        + dabsQc*dphi*cov[2][3]
                    )
                )
                self.fitresults.update({
                        "fr_err": fr_err,
                        "Ql_err": Ql_err,
                        "absQc_err": absQc_err,
                        "phi_err": phi_err,
                        "Qi_err": Qi_err,
                        "Qi_no_dia_corr_err": Qi_no_dia_corr_err,
                        "chi_square": chi_square
                })
            else:
                logging.warning("Error calculation failed!")
        else:
            # Just calculate reduced chi square (4 fit parameters reduce degrees
            # of freedom)
            self.fitresults["chi_square"] = (1. / (len(self.f_data) - 4.)
                * np.sum(np.abs(self._get_residuals_reflection)**2))
                
    def calc_fano_range(self, isolation=15, b=None):
        """
        Calculates the systematic Qi (and Qc) uncertainty range based on
        Fano interference with given strength of the background path
        (cf. Rieger & Guenzler et al., arXiv:2209.03036).
        
        inputs: either of
        - isolation (dB): Suppression of the interference path by this value.
                          The corresponding relative background amplitude b
                          is calculated with b = 10**(-isolation/20).
        - b (lin): Relative background path amplitude of Fano.

        outputs (added to fitresults dictionary):
        - Qi_min, Qi_max: Systematic uncertainty range for Qi
        - Qc_min, Qc_max: Systematic uncertainty range for Qc
        - fano_b: Relative background path amplitude of Fano.
        """
        
        if b is None:
            b = 10**(-isolation/20)
            
        b = b / (1 - b)
        
        if np.sin(self.phi) > b:
            logging.warning(
                "Measurement cannot be explained with assumed Fano leakage!"
            )
            self.Qi_min = np.nan
            self.Qi_max = np.nan
            self.Qc_min = np.nan
            self.Qc_max = np.nan
        
        # Calculate error on radius of circle
        R_mid = self.r0 * np.cos(self.phi)
        R_err = self.r0 * np.sqrt(b**2 - np.sin(self.phi)**2)
        R_min = R_mid - R_err
        R_max = R_mid + R_err
        
        # Convert to ranges of quality factors
        self.Qc_min = self.Ql / (self.n_ports*R_max)
        self.Qc_max = self.Ql / (self.n_ports*R_min)
        self.Qi_min = self.Ql / (1 - self.n_ports*R_min)
        self.Qi_max = self.Ql / (1 - self.n_ports*R_max)
        
        # Handle unphysical results
        if R_max >= 1./self.n_ports:
            self.Qi_max = np.inf
        
        # Store results in dictionary
        self.fitresults.update({
            "Qc_min": self.Qc_min,
            "Qc_max": self.Qc_max,
            "Qi_min": self.Qi_min,
            "Qi_max": self.Qi_max,
            "fano_b": b
        })
    
    def _fit_circle(self, z_data, refine_results=False):
        """
        Analytical fit of a circle to  the scattering data z_data. Cf. Sebastian
        Probst: "Efficient and robust analysis of complex scattering data under
        noise in microwave resonators" (arXiv:1410.3365v2)
        """
        
        # Normalize circle to deal with comparable numbers
        x_norm = 0.5*(np.max(z_data.real) + np.min(z_data.real))
        y_norm = 0.5*(np.max(z_data.imag) + np.min(z_data.imag))
        z_data = z_data[:] - (x_norm + 1j*y_norm)
        amp_norm = np.max(np.abs(z_data))
        z_data = z_data / amp_norm
        
        # Calculate matrix of moments
        xi = z_data.real
        xi_sqr = xi*xi
        yi = z_data.imag
        yi_sqr = yi*yi
        zi = xi_sqr+yi_sqr
        Nd = float(len(xi))
        xi_sum = xi.sum()
        yi_sum = yi.sum()
        zi_sum = zi.sum()
        xiyi_sum = (xi*yi).sum()
        xizi_sum = (xi*zi).sum()
        yizi_sum = (yi*zi).sum()
        M =  np.array([
            [(zi*zi).sum(), xizi_sum, yizi_sum, zi_sum],
            [xizi_sum, xi_sqr.sum(), xiyi_sum, xi_sum],
            [yizi_sum, xiyi_sum, yi_sqr.sum(), yi_sum],
            [zi_sum, xi_sum, yi_sum, Nd]
        ])
    
        # Lets skip line breaking at 80 characters for a moment :D
        a0 = ((M[2][0]*M[3][2]-M[2][2]*M[3][0])*M[1][1]-M[1][2]*M[2][0]*M[3][1]-M[1][0]*M[2][1]*M[3][2]+M[1][0]*M[2][2]*M[3][1]+M[1][2]*M[2][1]*M[3][0])*M[0][3]+(M[0][2]*M[2][3]*M[3][0]-M[0][2]*M[2][0]*M[3][3]+M[0][0]*M[2][2]*M[3][3]-M[0][0]*M[2][3]*M[3][2])*M[1][1]+(M[0][1]*M[1][3]*M[3][0]-M[0][1]*M[1][0]*M[3][3]-M[0][0]*M[1][3]*M[3][1])*M[2][2]+(-M[0][1]*M[1][2]*M[2][3]-M[0][2]*M[1][3]*M[2][1])*M[3][0]+((M[2][3]*M[3][1]-M[2][1]*M[3][3])*M[1][2]+M[2][1]*M[3][2]*M[1][3])*M[0][0]+(M[1][0]*M[2][3]*M[3][2]+M[2][0]*(M[1][2]*M[3][3]-M[1][3]*M[3][2]))*M[0][1]+((M[2][1]*M[3][3]-M[2][3]*M[3][1])*M[1][0]+M[1][3]*M[2][0]*M[3][1])*M[0][2]
        a1 = (((M[3][0]-2.*M[2][2])*M[1][1]-M[1][0]*M[3][1]+M[2][2]*M[3][0]+2.*M[1][2]*M[2][1]-M[2][0]*M[3][2])*M[0][3]+(2.*M[2][0]*M[3][2]-M[0][0]*M[3][3]-2.*M[2][2]*M[3][0]+2.*M[0][2]*M[2][3])*M[1][1]+(-M[0][0]*M[3][3]+2.*M[0][1]*M[1][3]+2.*M[1][0]*M[3][1])*M[2][2]+(-M[0][1]*M[1][3]+2.*M[1][2]*M[2][1]-M[0][2]*M[2][3])*M[3][0]+(M[1][3]*M[3][1]+M[2][3]*M[3][2])*M[0][0]+(M[1][0]*M[3][3]-2.*M[1][2]*M[2][3])*M[0][1]+(M[2][0]*M[3][3]-2.*M[1][3]*M[2][1])*M[0][2]-2.*M[1][2]*M[2][0]*M[3][1]-2.*M[1][0]*M[2][1]*M[3][2])
        a2 = ((2.*M[1][1]-M[3][0]+2.*M[2][2])*M[0][3]+(2.*M[3][0]-4.*M[2][2])*M[1][1]-2.*M[2][0]*M[3][2]+2.*M[2][2]*M[3][0]+M[0][0]*M[3][3]+4.*M[1][2]*M[2][1]-2.*M[0][1]*M[1][3]-2.*M[1][0]*M[3][1]-2.*M[0][2]*M[2][3])
        a3 = (-2.*M[3][0]+4.*M[1][1]+4.*M[2][2]-2.*M[0][3])
        a4 = -4.
    
        def char_pol(x):
            return a0 + a1*x + a2*x**2 + a3*x**3 + a4*x**4
    
        def d_char_pol(x):
            return a1 + 2*a2*x + 3*a3*x**2 + 4*a4*x**3
    
        eta = spopt.newton(char_pol, 0., fprime=d_char_pol)
    
        M[3][0] = M[3][0] + 2*eta
        M[0][3] = M[0][3] + 2*eta
        M[1][1] = M[1][1] - eta
        M[2][2] = M[2][2] - eta
        
        U,s,Vt = np.linalg.svd(M)
        A_vec = Vt[np.argmin(s),:]
    
        xc = -A_vec[1]/(2.*A_vec[0])
        yc = -A_vec[2]/(2.*A_vec[0])
        # The term *sqrt term corrects for the constraint, because it may be
        # altered due to numerical inaccuracies during calculation
        r0 = 1./(2.*np.absolute(A_vec[0]))*np.sqrt(
            A_vec[1]*A_vec[1]+A_vec[2]*A_vec[2]-4.*A_vec[0]*A_vec[3]
        )
        
        return xc*amp_norm+x_norm, yc*amp_norm+y_norm, r0*amp_norm
    
    def _fit_phase(self, z_data, guesses=None):
        """
        Fits the phase response of a strongly overcoupled (Qi >> Qc) resonator
        in reflection which corresponds to a circle centered around the origin
        (cf‌. phase_centered()).

        inputs:
        - z_data: Scattering data of which the phase should be fit. Data must be
                  distributed around origin ("circle-like").
        - guesses (opt.): If not given, initial guesses for the fit parameters
                          will be determined. If given, should contain useful
                          guesses for fit parameters as a tuple (fr, Ql, delay)

        outputs:
        - fr: Resonance frequency
        - Ql: Loaded quality factor
        - theta: Offset phase
        - delay: Time delay between output and input signal leading to linearly
                 frequency dependent phase shift
        """
        phase = np.unwrap(np.angle(z_data))
        
        # For centered circle roll-off should be close to 2pi. If not warn user.
        if np.max(phase) - np.min(phase) <= 0.8*2*np.pi:
            logging.warning(
                "Data does not cover a full circle (only {:.1f}".format(
                    np.max(phase) - np.min(phase)
                )
               +" rad). Increase the frequency span around the resonance?"
            )
            roll_off = np.max(phase) - np.min(phase)
        else:
            roll_off = 2*np.pi
        
        # Set useful starting parameters
        if guesses is None:
            # Use maximum of derivative of phase as guess for fr
            phase_smooth = gaussian_filter1d(phase, 30)
            phase_derivative = np.gradient(phase_smooth)
            fr_guess = self.f_data[np.argmax(np.abs(phase_derivative))]
            Ql_guess = 2*fr_guess / (self.f_data[-1] - self.f_data[0])
            # Estimate delay from background slope of phase (substract roll-off)
            slope = phase[-1] - phase[0] + roll_off
            delay_guess = -slope / (2*np.pi*(self.f_data[-1]-self.f_data[0]))
        else:
            fr_guess, Ql_guess, delay_guess = guesses
        # This one seems stable and we do not need a manual guess for it
        theta_guess = 0.5*(np.mean(phase[:5]) + np.mean(phase[-5:]))
        
        # Fit model with less parameters first to improve stability of fit
        
        def residuals_Ql(params):
            Ql, = params
            return residuals_full((fr_guess, Ql, theta_guess, delay_guess))
        def residuals_fr_theta(params):
            fr, theta = params
            return residuals_full((fr, Ql_guess, theta, delay_guess))
        def residuals_delay(params):
            delay, = params
            return residuals_full((fr_guess, Ql_guess, theta_guess, delay))
        def residuals_fr_Ql(params):
            fr, Ql = params
            return residuals_full((fr, Ql, theta_guess, delay_guess))
        def residuals_full(params):
            return self._phase_dist(
                phase - circuit.phase_centered(self.f_data, *params)
            )

        p_final = spopt.leastsq(residuals_Ql, [Ql_guess])
        Ql_guess, = p_final[0]
        p_final = spopt.leastsq(residuals_fr_theta, [fr_guess, theta_guess])
        fr_guess, theta_guess = p_final[0]
        p_final = spopt.leastsq(residuals_delay, [delay_guess])
        delay_guess, = p_final[0]
        p_final = spopt.leastsq(residuals_fr_Ql, [fr_guess, Ql_guess])
        fr_guess, Ql_guess = p_final[0]
        p_final = spopt.leastsq(residuals_full, [
            fr_guess, Ql_guess, theta_guess, delay_guess
        ])
        
        return p_final[0]
        
    @classmethod
    def phase_centered(cls, f, fr, Ql, theta, delay=0.):
        """
        Yields the phase response of a strongly overcoupled (Qi >> Qc) resonator
        in reflection which corresponds to a circle centered around the origin.
        Additionally, a linear background slope is accounted for if needed.
        
        inputs:
        - fr: Resonance frequency
        - Ql: Loaded quality factor (and since Qi >> Qc also Ql = Qc)
        - theta: Offset phase
        - delay (opt.): Time delay between output and input signal leading to
                        linearly frequency dependent phase shift
        """
        return theta - 2*np.pi*delay*(f-fr) + 2.*np.arctan(2.*Ql*(1. - f/fr))
    
    def _phase_dist(self, angle):
        """
        Maps angle [-2pi, +2pi] to phase distance on circle [0, pi]
        """
        return np.pi - np.abs(np.pi - np.abs(angle))
        
    def _periodic_boundary(self, angle):
        """
        Maps arbitrary angle to interval [-np.pi, np.pi)
        """
        return (angle + np.pi) % (2*np.pi) - np.pi
    
    def _get_residuals(self):
        """
        Calculates deviation of measured data from fit.
        """
        return self.z_data_norm - self.Sij(
            self.f_data, self.fr, self.Ql, self.Qc, self.phi
        )
    
    def _get_covariance(self):
        """
        Calculates reduced chi square and covariance matrix for fit.
        """
        residuals = self._get_residuals()
        chi = np.abs(residuals)
        # Unit vectors pointing in the correct directions for the derivative
        directions = residuals / chi
        # Prepare for fast construction of Jacobian
        conj_directions = np.conj(directions) 
    
        # Construct transpose of Jacobian matrix
        Jt = np.array([
            np.real(self._dSij_dfr()*conj_directions),
            np.real(self._dSij_dQl()*conj_directions),
            np.real(self._dSij_dabsQc()*conj_directions),
            np.real(self._dSij_dphi()*conj_directions)
        ])
        A = np.dot(Jt, np.transpose(Jt))
        # 4 fit parameters reduce degrees of freedom for reduced chi square
        chi_square = 1./float(len(self.f_data)-4) * np.sum(chi**2)
        try:
            cov = np.linalg.inv(A)*chi_square
        except:
            cov = None
        return chi_square, cov
    
    def _dSij_dfr(self):
        """
        Derivative of Sij w.r.t. fr
        """
        return -4j*self.Ql**2*np.exp(1j*self.phi)*self.f_data / (
            self.n_ports * self.absQc*(self.fr+2j*self.Ql*(self.f_data-self.fr))**2
        )
        
    def _dSij_dQl(self):
        """
        Derivative of Sij w.r.t. Ql
        """
        return -2.*np.exp(1j*self.phi) / (
            self.n_ports * self.absQc*(1.+2j*self.Ql*(self.f_data/self.fr-1))**2
        )
        
    def _dSij_dabsQc(self):
        """
        Derivative of Sij w.r.t. absQc
        """
        return 2.*self.Ql*np.exp(1j*self.phi) / (
            self.n_ports * self.absQc**2 * (1.+2j*self.Ql*(self.f_data/self.fr-1))
        )
        
    def _dSij_dphi(self):
        """
        Derivative of Sij w.r.t. phi
        """
        return -2j*self.Ql*np.exp(1j*self.phi) / (
            self.n_ports * self.absQc * (1.+2j*self.Ql*(self.f_data/self.fr-1))
        )
        
    """
    Functions for plotting results
    """
    def plotall(self):
        if not plot_enable:
            raise ImportError("matplotlib not found")
        real = self.z_data_raw.real
        imag = self.z_data_raw.imag
        real2 = self.z_data_sim.real
        imag2 = self.z_data_sim.imag
        plt.subplot(221, aspect="equal")
        plt.axvline(0, c="k", ls="--", lw=1)
        plt.axhline(0, c="k", ls="--", lw=1)
        plt.plot(real,imag,label='rawdata')
        plt.plot(real2,imag2,label='fit')
        plt.xlabel('Re(S21)')
        plt.ylabel('Im(S21)')
        plt.legend()
        plt.subplot(222)
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data_raw),label='rawdata')
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data_sim),label='fit')
        plt.xlabel('f (GHz)')
        plt.ylabel('|S21|')
        plt.legend()
        plt.subplot(223)
        plt.plot(self.f_data*1e-9,np.angle(self.z_data_raw),label='rawdata')
        plt.plot(self.f_data*1e-9,np.angle(self.z_data_sim),label='fit')
        plt.xlabel('f (GHz)')
        plt.ylabel('arg(|S21|)')
        plt.legend()
        plt.show()
        
    def plotcalibrateddata(self):
        if not plot_enable:
            raise ImportError("matplotlib not found")
        real = self.z_data_norm.real
        imag = self.z_data_norm.imag
        plt.subplot(221)
        plt.plot(real,imag,label='rawdata')
        plt.xlabel('Re(S21)')
        plt.ylabel('Im(S21)')
        plt.legend()
        plt.subplot(222)
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data_norm),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('|S21|')
        plt.legend()
        plt.subplot(223)
        plt.plot(self.f_data*1e-9,np.angle(self.z_data_norm),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('arg(|S21|)')
        plt.legend()
        plt.show()
        
    def plotrawdata(self):
        if not plot_enable:
            raise ImportError("matplotlib not found")
        real = self.z_data_raw.real
        imag = self.z_data_raw.imag
        plt.subplot(221)
        plt.plot(real,imag,label='rawdata')
        plt.xlabel('Re(S21)')
        plt.ylabel('Im(S21)')
        plt.legend()
        plt.subplot(222)
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data_raw),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('|S21|')
        plt.legend()
        plt.subplot(223)
        plt.plot(self.f_data*1e-9,np.angle(self.z_data_raw),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('arg(|S21|)')
        plt.legend()
        plt.show()
    
class reflection_port(circuit):
    """
    Circlefit class for single-port resonator probed in reflection.
    """
    
    # See Sij of circuit class for explanation
    n_ports = 1.
    
class notch_port(circuit):
    """
    Circlefit class for two-port resonator probed in transmission.
    """
    
    # See Sij of circuit class for explanation
    n_ports = 2.


# Driver for a StepMotor to rotate a phase shifter
# @KIT 2018 TW
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import zerorpc
from qkit.core.instrument_base import Instrument
import logging
import types
from qkit.gui.notebook.Progress_Bar import Progress_Bar


class StepMotor(Instrument):
    """
    Driver for the step motor connected to a raspberry pi. Connection from PC to raspberry pi uses zerorpc.
    Driver tracks the rotation of the phase shifter to not cause any damage. That means you have to provide it's
    initial position in rotations (float). Phase shifter can have a max of 17.5 rotations. Counting positive in the
    mathematical sense if you look onto the screw from the motor's perspective. (Between motor and phase shifter is
    a 27:1 gearbox. This means rotations are rather slow.)
    """

    def __init__(self, name, address="tcp://10.22.197.112:4242", initial_position=0):

        self.c = zerorpc.Client()
        self.c.connect(address)

        if self.c.get_connection():
            print 'connected to Raspberry Pi'
        else:
            logging.error('No connection to Raspberry Pi')

        if initial_position == 0:
            logging.warning("You didn't provide an inital position. Please make sure that the phase shifter "
                            "is in its lowest position, since the rotation is tracked to prevent the damage of "
                            "either motor or phase shifter. Or alternatively change the rotations."
                            "Please also think of that when manually changing the phase shifter")
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self.angle = initial_position * 360
        self.min_angle = 0
        self.max_angle = 17.5 * 360
        self.rotations = initial_position
        self.show_progress_bar = True

        self.add_parameter('angle', type=int, flags=Instrument.FLAG_GETSET, minval=0, maxval=17.5 * 360)
        self.add_parameter('waiting_time', type=float, flags=Instrument.FLAG_GETSET,
                           minval=0.0001, maxval=0.01)
        self.add_parameter('rotations', type=float, flags=Instrument.FLAG_GETSET, minval=0, maxval=17.5 * 360)
        self.add_function('turn')

    def do_get_angle(self):
        """
        :return: the current angle of the phase shifter. (Works only if the phase shifter was in the "zero"-state
                at the beginning
        """
        return self.angle

    def do_set_angle(self, angle):
        """
        Sets the phase shifter to a given value.
        :param angle:
        :return: None
        """
        difference = angle - self.angle
        self.c.turn(difference)

    def turn(self, angle):
        """
        Turns the step motor and thus the phase shifter at a given angle
        :param angle: value in degrees positive or negative.
        :return: None
        """
        self.angle += angle
        if self.angle < self.min_angle or self.angle > self.max_angle:
            logging.error("Out of range, operation aborted to prevent damage!")
            self.angle -= angle
        else:
            self.c.turn(angle)
            max_steps = self.c.get_max_steps()
            if self.show_progress_bar:
                pb = Progress_Bar(max_steps - 2)
            try:
                while self.c.get_steps() < max_steps - 1:
                    if self.show_progress_bar:
                        for i in range(self.c.get_steps() - pb.progr):
                            pb.iterate()
                    else:
                        pass
            except KeyboardInterrupt:
                logging.warning("KeyboardInterrupt: Rotation stopped!")
                self.c.stop_rotation()
                self.angle -= int(angle * (float(max_steps - self.c.get_steps())) / max_steps)

    def do_set_waiting_time(self, waiting_time):
        """
        This is a function to specify the speed of your step motor. Please handle with care. The shorter
        the waiting time between steps the faster is your step motor (default = 0.0001 sec)
        :param waiting_time: time in sec between 0.0001 and 0.01
        :return:
        """
        self.c.set_waiting_time(waiting_time)

    def do_get_waiting_time(self):
        """
        :return: the waiting time between steps and thus the speed of your step motor
        """
        return self.c.get_waiting_time()

    def do_get_rotations(self):
        """
        :return: the current position of the phase shifter if correctly initialized
        """
        self.rotations = float(self.angle) / 360
        return self.rotations

    def do_set_rotations(self, position):
        """
        Used to set the position of the phase shifter if manually changed in rotations from lowest postion.
        positive rotation in the mathematical sense when looking from motor perspective
        :param position: number of rotations of the phase shifter from zero (float)
        :return: None
        """
        self.rotations = position



from abc import ABC, abstractmethod
from typing import Union
from typing_extensions import Self
import black
from matplotlib.pyplot import isinteractive

import numpy as np

from zhinst.toolkit.control.node_tree import NodeList
from qkit.core.instrument_base import Instrument
SETTING_TAG = 'setting'

class ZHInst_Abstract(ABC, Instrument):

    def __init__(self, name, **kwargs) -> None:
        super().__init__(name, **kwargs)
        if not hasattr(self, "blacklist"):
            self.blacklist = []
        self.add_tag(SETTING_TAG)

    def mount_api(self, file_name: str, node_tree):
        with open(file_name, "w", encoding="utf-8") as f:
            visitor = self.NodeTreeVisitor(f, self, self.blacklist)
            visitor.visit_node([node_tree])

    class NodeTreeVisitor:

        def __init__(self, doc_file, target, blacklist) -> None:
            self.doc_file = doc_file
            self.target = target
            self.blacklist = blacklist


        def visit_parameter(self, parameters: list, path=[]):
            prefix, func_name = self._derive_channel_affix(path)
            if any(map(lambda it: func_name.startswith(it), self.blacklist)):
                return
            parameters = list(parameters)
            access = parameters[0]._properties # A set of strings describing access: Read, Write, Setting
            if 'Setting' in access:
                tags = [SETTING_TAG] 
            else:
                tags = []

            def get_call(zhinst_getters=parameters, channel=0): # We need the kv-arg to capture the value.
                return zhinst_getters[channel]()

            try:
                get_call()
            except Exception as e:
                print(f"Can't access {func_name}.")
                print(e)
                return

            setattr(self.target, f"_do_get_{func_name}", get_call)
            if 'Write' in access:
                def set_call(value, zhinst_setters=parameters, channel=0):
                    zhinst_setters[channel](value)
                setattr(self.target, f"_do_set_{func_name}", set_call)
                access = Instrument.FLAG_GETSET
            else:
                access = Instrument.FLAG_GET
            property_type = parameters[0]._type

            if property_type == 'Double':
                property_type = float
            elif property_type == 'Integer' or property_type == 'Integer (enumerated)' or property_type == 'Integer (64 bit)':
                property_type = int
            elif property_type == 'String':
                property_type = str
            elif property_type == 'ZIVectorData':
                property_type = np.ndarray
            else:
                import sys
                print(f"Type '{property_type}' of '{func_name}' is not supported!", file=sys.stderr)
                return

            unit = getattr(parameters[0], "_unit", None)
            
            if len(parameters) > 1:
                self.target.add_parameter(func_name, channel_prefix = prefix, type = property_type, flags = access, units = unit, doc = parameters.__repr__(), tags = tags, channels = (0, len(parameters) - 1))
            else:
                self.target.add_parameter(func_name, type = property_type, flags = access, units = unit, doc = parameters.__repr__(), tags = tags)
            # Lastly, for user reference, we print a file containing all mappings.
            # This will later be a reference manual, so that we can use meta-programming
            # to create all hooks, adapt to future updates, and let the user know what to call.
            print(f"get/set_{func_name}:", file=self.doc_file)
            print(parameters[0].__repr__(), file=self.doc_file)

        def visit_node(self, nodes: list, path=[]):
            if self._check_all_have_non_deviating(nodes, lambda node: hasattr(node, "parameters")):
                for param_name in nodes[0].parameters:
                    self.visit_parameter(self._map_to_list(lambda node: getattr(node, param_name), nodes), path = path + [param_name])
            
            if self._check_all_have_non_deviating(nodes, lambda node: hasattr(node, "nodes")):
                for node_name in nodes[0].nodes:
                    if node_name == "_parent":
                        continue
                    self.visit_node(self._map_to_list(lambda node: getattr(node, node_name), nodes), path = path + [node_name])
            
            if self._check_all_have_non_deviating(nodes, lambda node: isinstance(node, NodeList)):
                if len(nodes) == 1: # We have not yet hit such a split -> Channel Marker
                    marker = self.ChannelNodeMarker(0, len(nodes[0]))
                    self.visit_node([n for n in nodes[0]], path = path + [marker])
                else: # Go down recursively, each iteration the next index
                    for i in range(len(nodes[0])):
                        self.visit_node(self._map_to_list(lambda node: node[i], nodes), path = path + [str(i)])

        
        def _derive_channel_affix(self, path):
            splits = [i for i, x in enumerate(path) if isinstance(x, self.ChannelNodeMarker)]
            if len(splits) > 0:
                split_index = splits[0]
                prefix = path[split_index - 1] + "_%s_"
                name = self._derive_channel_name(path[:split_index] + path[split_index + 1:])
                return  prefix, name
            else:
                return "", self._derive_channel_name(path)

        @staticmethod
        def _map_to_list(expr, l):
            return list(map(expr, l))

        @staticmethod
        def _derive_channel_name(path):
            return "_".join(map(lambda it: str(it), path))

        @staticmethod
        def _check_all_have_non_deviating(nodes, expression):
            return all(map(expression, nodes)) and not any(map(lambda node: not expression(node), nodes))
        
        class ChannelNodeMarker:
            
            lower_index_bound: int
            upper_index_bound: int

            def __init__(self, lower_index, upper_index) -> None:
                self.lower_index_bound = lower_index
                self.upper_index_bound = upper_index

            def __repr__(self) -> str:
                return f"{self.lower_index_bound}..{self.upper_index_bound}"

class ZHInst_AWG(ABC):
    """
    This is a mixin for all Zurich Instruments Devices which support an AWG.
    """


import numpy as np


def const(amplitude, length):
    const_wf = amplitude * np.ones(length)
    return const_wf

def const_zeroed(amplitude, zeros_start, zeros_end, length):

    const_zero_padded_wave = amplitude * np.ones(length)
    const_zero_padded_wave[:zeros_start] = 0.0
    const_zero_padded_wave[- zeros_end:] = 0.0

    return const_zero_padded_wave

def gauss(amplitude, delta_t, sigma, length):

    t = np.linspace(-length / 2, length / 2, length)
    gauss_wave = amplitude * np.exp(-((t - delta_t) ** 2) / (2 * sigma ** 2))

    return gauss_wave

def plot_pulses():

    pass






import qkit

# Check if we are using the new data structure and if we have set user and RunID
if 'new_data_structure' in qkit.cfg:
    raise ValueError(__name__+": Please use qkit.cfg['datafolder_structure'] = 1 instead of qkit.cfg['new_data_structure'] in your config.")
if qkit.cfg.get('datafolder_structure', 1) == 2:
    # noinspection SpellCheckingInspection
    try:
        import ipywidgets as widgets
        from IPython.display import display
        
        b = widgets.Button(
                description='Please Check!',
                disabled=False,
                button_style='info',  # 'success', 'info', 'warning', 'danger' or ''
        )
        
        b.f1 = widgets.Text(
                value=str(qkit.cfg.get('run_id', '')).upper(),
                placeholder='***RUN_ID IS EMPTY***',
                description='Please check: Run ID',
                disabled=False,
                style={'description_width': 'initial'}
        )
        
        b.f2 = widgets.Text(
                value=str(qkit.cfg.get('user', '')),
                placeholder='***USER IS EMPTY***',
                description='user name',
                disabled=False,
                style={'description_width': 'initial'}
        )
        if not qkit.cfg.get('run_id', False):
            b.f1.border_color = 'red'
            b.button_style = 'danger'
        
        if not qkit.cfg.get('user', False):
            b.f2.border_color = 'red'
            b.button_style = 'danger'
        
        
        def clickfunc(btn):
            if not b.f1.value:
                raise ValueError("RUN_ID is still empty!")
            if not b.f2.value:
                raise ValueError("USER is still empty!")
            qkit.cfg['run_id'] = b.f1.value.upper()
            qkit.cfg['user'] = b.f2.value
            btn.f1.disabled = True  # close()
            btn.f1.border_color = '#cccccc'
            btn.f2.border_color = '#cccccc'
            btn.f2.disabled = True  # close()
            btn.disabled = True  # ()
            btn.button_style = 'success'
            btn.description = 'Done.'
        
        
        b.on_click(clickfunc)
        display(widgets.HBox([b.f1, b.f2, b]))
    except ImportError:
        import logging
        
        if 'run_id' not in qkit.cfg:
            logging.error(
                    'You are using the new data structure, but you did not specify a run ID. Please set qkit.cfg["run_id"] NOW to avoid searching your data.')
        if 'user' not in qkit.cfg:
            logging.error(
                    'You are using the new data structure, but you did not specify a username. Please set qkit.cfg["user"] NOW to avoid searching your data.')



# root init for QKIT
# these directories are treated as modules
# HR@KIT/2017/2018
__all__ = ['config','gui','measure','tools', 'analysis','core','instruments','services','storage','logs']

import os.path
import logging
class QkitCfgError(Exception):
    '''
    If something with qkit.cfg does not fit to what the user wants to do, this is the error to throw.
    '''
    pass

class ConfClass(dict):
    def __init__(self, *args):
        dict.__init__(self, args)
    def preset_analyse(self,verbose = False):
        """ Sets basic settings, most of the services are not loaded (default)
            The file index service is run and the UUID registry is populated.
        """

        self['load_info_service'] = False
        self['load_ri_service']   = False
        self['load_visa']         = False
        if verbose:
            print ("Not starting the info_service, ri_service and visa.")
    def preset_measure(self,verbose = False):
        """ Setup of the measurement settings, services are loaded or initialized.
        """
        self['load_info_service'] = True
        self['load_ri_service']   = True
        self['load_visa']         = True
        
        if self['datadir'] == os.path.join(self['qkitdir'],'data'):
                print("Please set a valid data directory! (datadir)")
        if verbose:
            print ("Starting the info_service, ri_service and visa.")
            
    def get(self, item,default=None):
        try:
            return self[item]
        except KeyError:
            if default is not None:
                self[item]=default
            return default
cfg = ConfClass()

# load configuration from $QKITDIR/config/*

try:
    from qkit.config.environment import cfg as cfg_local
    cfg.update(cfg_local)
except ImportError:
    pass

# if a local.py file is defined, load cfg dict and overwrite environment entries.
_config_sources = []
try:
    from qkit.config.local import cfg_local
    cfg.update(cfg_local)
    _config_sources += ["[qkit]src/qkit/config/local.py - cfg_local"]
    logging.warning("DEPRECATED: Loaded qkit.config.local.py! This has been deprecated in favour of qkit_local_config.py, see README!")
except ImportError:
    pass

try:    
    from qkit.config.local import cfg as cfg_local
    cfg.update(cfg_local)
    _config_sources += ["[qkit]src/qkit/config/local.py - cfg"]
    logging.warning("DEPRECATED: Loaded qkit.config.local.py! This has been deprecated in favour of qkit_local_config.py, see README!")
except ImportError:
    pass


def _load_user_config():
    """
    Loads user configuration from the file system, determined by known file names or environment variables.

    We usually run in a jupyter notebook. If qkit is installed as a package, a local config is not possible,
    and in the long run we should migrate towards a config not in the qkit directory.
    We will use two strategies here:
    1. Check if the environment variable 'QKIT_LOCAL_CONFIG' is set. Use the file it points to.
    2. Use current working directory:
      a. check if a qkit_conf.py file exists. If it does, use it.
      b. Else go to parrent directory and repeat
    If a file was found by either way, load it.

    We log information to stdout, as logging is not yet available. This code is executed before S10_logging.py,
    and will already have become unavailable once logging is initialized.
    """
    global _config_sources
    import os
    from pathlib import Path
    from itertools import chain

    user_config = None
    environment_variable_name = 'QKIT_LOCAL_CONFIG'
    config_file_names = ['qkit_local_config.py', 'local.py']

    if environment_variable_name in os.environ:
        # Inspect environment variable, should it exist
        print("Found environment variable '%s' for configuration." % environment_variable_name)
        user_config = os.environ[environment_variable_name]
    else:
        current_directory = Path(os.getcwd())
        for dir in chain([current_directory], current_directory.parents):
            for (index, config_file_name) in enumerate(config_file_names):
                test_path = dir / config_file_name
                if test_path.exists():
                    if test_path.is_file():
                        user_config = test_path
                        if index != 0:
                            print("WARNING: Found legacy configuration file name. Consider migrating.")
                        # We are done here
                        break
                    else:
                        print("WARNING: Found correct file system name %s, but was not a file?!" % test_path)

    # Load if found
    if user_config is not None:
        # https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
        import importlib.util
        spec = importlib.util.spec_from_file_location("userconfig", user_config)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        cfg.update(module.cfg)
        _config_sources += [user_config]

_load_user_config()
del _load_user_config

if len(_config_sources) > 1:
    logging.warn("Loaded more than one configuration file! This may cause conflicts! Sources:")
    logging.warn(_config_sources)
del _config_sources

# clean up 
del cfg_local
# init message
print ("QKIT configuration initialized -> available as qkit.cfg[...]")

"""
startup functions
"""
# start initialization (qkit/core/init)
def start(silent=False):
    if not silent:
        print("Starting QKIT framework ... -> qkit.core.startup")
    import qkit.core.startup
    qkit.core.startup.start(silent=silent)


"""
add a few convenience shortcuts 
"""
# remote interface client after qkit.start_ric() -> qkit.ric
from qkit.core.lib.com.ri_client import start_ric



# -*- coding: utf-8 -*-

# MMW@KIT
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
References
----------
https://pypi.org/project/pcf8575/

Installation
------------
sudo apt-get update
sudo apt-get install python-rpi.gpio
sudo apt-get install libffi-dev
pip install pcf8575
"""

import zmq
from zmq.auth.thread import ThreadAuthenticator
from qkit.config.services import cfg
import time
import logging
import os
import json

try:
    from pcf8575 import PCF8575
except ModuleNotFoundError:
    logging.warning('qplexkit: Module PCF8575 for I2C-expander not found. Use dummy instead that just supports software debugging but no control of the qplexkit at all.')
    class PCF8575(object):
        def __init__(self, i2c_bus_no, address):
            print(i2c_bus_no, address)
            self.port = 16 * [True]


class qplexkit(object):
    def __init__(self):
        self.pcf1 = PCF8575(i2c_bus_no=1, address=0x20)
        self.pcf2 = PCF8575(i2c_bus_no=1, address=0x21)
        self.pcf = [self.pcf1, self.pcf2]

    def set_port(self, pcf, port, status):
        self.pcf[pcf].port[port] = status
        #print(f'set port {port} to {status}')
        return

    def get_port(self, pcf, port):
        #print(f'get port {port}')
        return self.pcf[pcf].port[port]


qpk = qplexkit()

msg2cmd = {'set_port': qpk.set_port,
           'get_port': qpk.get_port,
           'get_attr': lambda attr, *args, **kwargs: getattr(qpk, attr),
           }


''' run zmq-server '''
if __name__ == "__main__":
    """
    -------
    Example
    >>> socket.send_json(("set_port", tuple([]), dict({'port': <port>, 'pcf': <pcf>, 'status': <status>})))
    >>> socket.recv_json()
    """
    context = zmq.Context()
    auth = ThreadAuthenticator(context, log=logging.getLogger())
    auth.start()
    auth.allow('localhost')
    auth.allow('10.22.197.0')
    #for ip in cfg['qplexkit']['allowed_ip_addresses'].split(', '):
    #    auth.allow(ip)
    socket = context.socket(zmq.REP)
    socket.zap_domain = b'global'
    socket.bind(f'''tcp://*:{cfg['qplexkit']['server_port']}''')

    while True:
        fun, args, kwargs = socket.recv_json()
        socket.send_json(msg2cmd[fun](*args, **kwargs))




# -*- coding: utf-8 -*-
"""
Remote interface client (RIC) to access the RI service or any zerorpc service


@author: HR,AS@KIT 2018

To use the client to control a qkit instance, just open a new notebook and execute the following:
    
    import qkit.core.lib.com.ri_client as rc
    rc.control_panel(rc.start_ric())
    
"""
import qkit

#import logging # the logging mechanism does not work here, since qkit is not started



def __update_tab(ric):
    # try to be smart and add the remote functions to the __dict__ for tabbing in ipython
    rflist = ric._zerorpc_list()
    for f in rflist: 
        ric.__dict__.update({f:getattr(ric,f)})

def __getdoc():
    return _zerorpc_help
    
    
def start_ric(host = None,port = None):
    import zerorpc
    """
    starts a remote interface client (ric)
    to a 
        host = host (default =  localhost)
    at a 
        port  = port (default =  5700)
    
    ric is based on zerorpc, so every zerorpc server should be accessible
    
    """
    ric = zerorpc.Client(); 
    if not host:
        host = qkit.cfg.get("ris_host","127.0.0.1")
    if not port:
        port = qkit.cfg.get("ris_port",5700)
    print("starting ric client on host %s and port %s" % (host,port))
    ric.connect("tcp://" + host + ":" + str(port))
    
    qkit.ric = ric    
    __update_tab(ric)
    
    return ric

"""
list of internal functions
_zerorpc_list to list calls
_zerorpc_name to know who you’re talking to
_zerorpc_ping (redundant with the previous one)
_zerorpc_help to retrieve the docstring of a call
_zerorpc_args to retrieve the argspec of a call
_zerorpc_inspect to retrieve everything at once
"""


def control_panel(ric = None):
    from ipywidgets import Button, HTML,HBox,Accordion
    from IPython.core.display import display
    def fmt(inp):
        if inp is None:
            return "--"
        elif type(inp) is float:
            return "{:.4g}".format(inp)
        else:
            return str(inp)

    if ric is None:
        ric = qkit.ric

    def update_instrument_params(b):
        insdict = ric.get_all_instrument_params()
        for ins in sorted(insdict):
            if not ins in b.accordions:
                b.accordions[ins] = Accordion()
            table = "<table style='line-height:180%'>"  # <tr style='font-weight:bold'><td> Parameter </td><td>Value</td></tr>
            for p in sorted(insdict[ins]):
                table += "<tr><td style='text-align:right;padding-right:10px'>" + str(p) + "</td><td>" + fmt(insdict[ins][p][0]) + insdict[ins][p][
                    1] + "</td></tr>"
            table += """</table>"""
            b.accordions[ins].children = [HTML(table)]
            b.accordions[ins].set_title(0, ins)
        for child in b.accordions.keys():
            if child not in insdict:
                del b.accordions[child]
        b.hbox.children = b.accordions.values()

    update_button = Button(description="Update")
    update_button.on_click(update_instrument_params)
    update_button.accordions = {}
    update_button.hbox = HBox()
    stop_button = Button(description="Stop measurement")
    stop_button.on_click(lambda b: qkit.ric.stop_measure())
    stop_button.button_style = "danger"
    update_instrument_params(update_button)
    display(HBox([update_button,stop_button]),update_button.hbox)
